#-*- coding: utf-8 -*-
"""
================================================================================

  This file is part of the AMUN source code, a program to perform Newtonian or
  relativistic magnetohydrodynamical simulations on uniform or adaptive grid.

  Copyright (C) 2021-2022 Grzegorz Kowal <grzegorz@amuncode.org>

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

================================================================================

 module: AMUN

  The base class for Amun's snapshot formats.

--------------------------------------------------------------------------------
"""
class Amun:
    """AMUN snapshot base class"""

    def __init__(self, path, version=-1):

        import os

        if os.path.exists(path):
            if os.path.isdir(path):
                self.path_is_file = False
                self.dirname      = path
                self.filename     = None
            else:
                self.path_is_file = True
                self.dirname      = os.path.dirname(path)
                self.filename     = os.path.basename(path)
        else:
            raise Exception("Path '{}' does not exist!".format(path))

        self.path        = path
        self.dataformat  = None
        self.version     = version
        self.chunkname   = '{:d}'
        self.attributes  = dict()
        self.chunks      = dict()
        self.variables   = dict()

        self.__init_snapshot__()
        self.__fill_attributes__()
        self.__fill_chunks__()
        self.__fill_variables__()

        self.__check_attributes__()
        self.__check_chunks__()
        self.__check_variables__()

        self.__complete_attributes__()
        self.__complete_chunks__()
        self.__complete_variables__()


    def __init_snapshot__(self):
        """
            Initializes the snapshot object by checking the file's data format
            and setting the field 'self.dataformat' (string).
        """
        pass


    def __fill_attributes__(self):
        """
            Read attributes from the snapshot file and add all attributed to
            the dictionary 'self.attributes'.
        """
        pass


    def __fill_chunks__(self):
        """
            Retrieve metadata about data blocks stored in the snapshot's chunks.
        """
        pass


    def __fill_variables__(self):
        """
            Retrieve the variable names from the snapshot file and add them
            to the list 'self.variables'.
        """
        pass


    def __check_attributes__(self):
        """
            Check the presence of mandatory attributes.
        """
        mandatory_attributes = [ 'isnap', 'nchunks', 'ndims', 'maxlev', \
                'nleafs', 'ncells', 'nghosts', 'xmin', 'ymin', 'zmin', 'xmax', \
                'ymax', 'zmax', 'xblocks', 'yblocks', 'zblocks' ]

        for attr in mandatory_attributes:
            if not attr in self.attributes:
                raise Exception("Mandatory attribute '{}' not set!".format(attr))


    def __check_chunks__(self):
        """
            Check the presence of required field in chunks.
        """
        if not len(self.chunks) >= 1:
            raise Exception("The snapshot {} has no chunks!".format(self.path))

        mandatory_fields = [ 'filename', 'dblocks', 'levels', 'bounds', 'coords' ]

        for nc in range(self.attributes['nchunks']):
            for field in mandatory_fields:
                if not field in self.chunks[nc]:
                    raise Exception("Mandatory field '{}' not set in chunk no. {}!".format(field, nc))


    def __check_variables__(self):
        """
            Check the presence of variables.
        """
        if len(self.variables) < 1:
            raise Exception("The snapshot {} has no variables stored!".format(self.path))


    def __complete_attributes__(self):
        """
            Add additional attributes.
        """
        if not 'toplev' in self.attributes:
            self.attributes['toplev'] = max([max(chunk['levels']) if chunk['dblocks'] > 0 else 1 for chunk in self.chunks.values()])

        if not 'xlen' in self.attributes:
            self.attributes['xlen'] = self.attributes['xmax'] - self.attributes['xmin']
        if not 'ylen' in self.attributes:
            self.attributes['ylen'] = self.attributes['ymax'] - self.attributes['ymin']
        if not 'zlen' in self.attributes:
            self.attributes['zlen'] = self.attributes['zmax'] - self.attributes['zmin']

        if not 'bcells' in self.attributes:
            self.attributes['bcells'] = self.attributes['ncells'] + 2 * self.attributes['nghosts']

        if not 'resistivity' in self.attributes:
            self.attributes['resistivity'] = 0
        if not 'viscosity' in self.attributes:
            self.attributes['viscosity'] = 0

        self.cell_size = dict()
        for l in range(self.attributes['maxlev']):
            self.cell_size[l+1] = self.attributes['xlen'] / (self.attributes['xblocks'] * self.attributes['ncells'] * 2**l)


    def __complete_chunks__(self):
        """
            Add additional attributes to chunks.
        """
        for n in range(len(self.chunks)):
            if not 'dims' in self.chunks[n]:
                self.chunks[n]['dims'] = [ self.attributes['bcells'] ]*self.attributes['ndims']
                self.chunks[n]['dims'][:0] = [ self.chunks[n]['dblocks'] ]


    def __complete_variables__(self):
        """
            Add derived variables.
        """
        self.variables['refinement level']                  = 'mlev'
        if 'dens' in self.variables:
            self.variables['density']                       = 'dens'
            self.variables['logarithm of density']          = 'logd'
        if 'pres' in self.variables:
            self.variables['pressure']                      = 'dens'
            self.variables['logarithm of pressure']         = 'logp'
        if all(v in self.variables for v in ['velx','vely','velz']):
            self.variables['velocity']                      = 'vvec'
            self.variables['velocity magnitude']            = 'velo'
            self.variables['x-velocity']                    = 'velx'
            self.variables['y-velocity']                    = 'vely'
            self.variables['z-velocity']                    = 'velz'
            self.variables['divergence of velocity']        = 'divv'
            self.variables['x-vorticity']                   = 'vorx'
            self.variables['y-vorticity']                   = 'vory'
            self.variables['z-vorticity']                   = 'vorz'
            self.variables['vorticity']                     = 'wvec'
            self.variables['vorticity magnitude']           = 'vort'
        if all(v in self.variables for v in ['magx','magy','magz']):
            self.variables['magnetic field']                = 'bvec'
            self.variables['magnetic field magnitude']      = 'magn'
            self.variables['x-magnetic field']              = 'magx'
            self.variables['y-magnetic field']              = 'magy'
            self.variables['z-magnetic field']              = 'magz'
            self.variables['divergence of magnetic field']  = 'divb'
            self.variables['x-current density']             = 'curx'
            self.variables['y-current density']             = 'cury'
            self.variables['z-current density']             = 'curz'
            self.variables['current density']               = 'jvec'
            self.variables['current density magnitude']     = 'curr'
        if 'bpsi' in self.variables:
            self.variables['magnetic divergence potential'] = 'bpsi'
        if 'pres' in self.variables and 'adiabatic_index' in self.attributes:
            self.variables['internal energy']               = 'eint'
        if all(v in self.variables for v in ['dens', 'pres']):
            self.variables['temperature']                   = 'temp'
        if self.attributes['eqsys'] in [ 'hd', 'mhd' ] \
                        and all(v in self.variables for v in ['dens','velx','vely','velz']):
            self.variables['kinetic energy']                = 'ekin'
        if self.attributes['eqsys'] in [ 'mhd', 'srmhd' ] \
                        and all(v in self.variables for v in ['magx','magy','magz']):
            self.variables['magnetic energy']               = 'emag'
            self.variables['magnetic pressure']             = 'emag'
        if all(v in self.variables for v in ['velx','vely','velz', 'magx','magy','magz']) in self.variables:
            self.variables['electric field']                = 'evec'
            self.variables['electric field magnitude']      = 'elec'
            self.variables['x-electric field']              = 'elex'
            self.variables['y-electric field']              = 'eley'
            self.variables['z-electric field']              = 'elez'
        if self.attributes['eqsys'] in [ 'srhd', 'srmhd' ] \
                        and all(v in self.variables for v in ['velx','vely','velz']):
            self.variables['Lorentz factor']                = 'lore'

        delete_datasets = [ 'dens', 'pres', 'velx', 'vely', 'velz', 'magx','magy','magz', 'bpsi' ]
        for v in delete_datasets:
            if v in self.variables:
                del self.variables[v]


    def __read_binary_data__(self, dataset_name, chunk_number):
        """
            Get dataset array from the given snapshot's chunk.
        """
        pass


    def __get_dataset__(self, dataset_name, chunk_number):
        """
            Get dataset array of name dataset_name from the file n.
        """
        import numpy

        dataset = self.variables[dataset_name]

        if dataset == 'mlev':
            dset = numpy.zeros(self.chunks[chunk_number]['dims'])
            for p in range(self.chunks[chunk_number]['dblocks']):
                dset[p,...] = self.chunks[chunk_number]['levels'][p]
        elif dataset == 'logd':
            dset = numpy.log10(self.__read_binary_data__('dens', chunk_number))
        elif dataset == 'logp':
            dset = numpy.log10(self.__read_binary_data__('pres', chunk_number))
        elif dataset == 'velo':
            tmp = self.__read_binary_data__('velx', chunk_number)
            dset = tmp**2
            tmp = self.__read_binary_data__('vely', chunk_number)
            dset += tmp**2
            tmp = self.__read_binary_data__('velz', chunk_number)
            dset += tmp**2
            dset = numpy.sqrt(dset)
        elif dataset == 'vvec':
            dset = [self.__read_binary_data__('velx', chunk_number), \
                    self.__read_binary_data__('vely', chunk_number), \
                    self.__read_binary_data__('velz', chunk_number)]
        elif dataset == 'magn':
            tmp = self.__read_binary_data__('magx', chunk_number)
            dset = tmp**2
            tmp = self.__read_binary_data__('magy', chunk_number)
            dset += tmp**2
            tmp = self.__read_binary_data__('magz', chunk_number)
            dset += tmp**2
            dset = numpy.sqrt(dset)
        elif dataset == 'bvec':
            dset = [self.__read_binary_data__('magx', chunk_number), \
                    self.__read_binary_data__('magy', chunk_number), \
                    self.__read_binary_data__('magz', chunk_number)]
        elif dataset == 'ekin':
            tmp = self.__read_binary_data__('velx', chunk_number)
            dset = tmp**2
            tmp = self.__read_binary_data__('vely', chunk_number)
            dset += tmp**2
            tmp = self.__read_binary_data__('velz', chunk_number)
            dset += tmp**2
            tmp = self.__read_binary_data__('dens', chunk_number)
            dset *= tmp
            dset *= 0.5
        elif dataset == 'emag':
            tmp = self.__read_binary_data__('magx', chunk_number)
            dset = tmp**2
            tmp = self.__read_binary_data__('magy', chunk_number)
            dset += tmp**2
            tmp = self.__read_binary_data__('magz', chunk_number)
            dset += tmp**2
            dset *= 0.5
        elif dataset == 'elex':
            by = self.__read_binary_data__('magy', chunk_number)
            bz = self.__read_binary_data__('magz', chunk_number)
            vy = self.__read_binary_data__('vely', chunk_number)
            vz = self.__read_binary_data__('velz', chunk_number)
            dset = vy * bz - vz * by
            if self.attributes['resistivity'] > 0:
                if self.attributes['ndims'] == 3:
                    tmp  = (numpy.roll(by,  1, axis=1) - numpy.roll(by, -1, axis=1))
                    tmp += (numpy.roll(bz, -1, axis=2) - numpy.roll(bz,  1, axis=2))
                else:
                    tmp  = (numpy.roll(bz, -1, axis=1) - numpy.roll(bz,  1, axis=1))

                for p in range(self.chunks[chunk_number]['dblocks']):
                    tmp[p,...] /= self.cell_size[self.chunks[chunk_number]['levels'][p]]
                dset -= 0.5 * self.attributes['resistivity'] * tmp
        elif dataset == 'eley':
            bx = self.__read_binary_data__('magx', chunk_number)
            bz = self.__read_binary_data__('magz', chunk_number)
            vx = self.__read_binary_data__('velx', chunk_number)
            vz = self.__read_binary_data__('velz', chunk_number)
            dset = vz * bx - vx * bz
            if self.attributes['resistivity'] > 0:
                if self.attributes['ndims'] == 3:
                    tmp  = (numpy.roll(bx, -1, axis=1) - numpy.roll(bx,  1, axis=1))
                    tmp += (numpy.roll(bz,  1, axis=3) - numpy.roll(bz, -1, axis=3))
                else:
                    tmp  = (numpy.roll(bz,  1, axis=2) - numpy.roll(bz, -1, axis=2))

                for p in range(self.chunks[chunk_number]['dblocks']):
                    tmp[p,...] /= self.cell_size[self.chunks[chunk_number]['levels'][p]]
                dset -= 0.5 * self.attributes['resistivity'] * tmp
        elif dataset == 'elez':
            bx = self.__read_binary_data__('magx', chunk_number)
            by = self.__read_binary_data__('magy', chunk_number)
            vx = self.__read_binary_data__('velx', chunk_number)
            vy = self.__read_binary_data__('vely', chunk_number)
            dset = vx * by - vy * bx
            if self.attributes['resistivity'] > 0:
                if self.attributes['ndims'] == 3:
                    tmp  = (numpy.roll(bx,  1, axis=2) - numpy.roll(bx, -1, axis=2))
                    tmp += (numpy.roll(by, -1, axis=3) - numpy.roll(by,  1, axis=3))
                else:
                    tmp  = (numpy.roll(bx,  1, axis=1) - numpy.roll(bx, -1, axis=1))
                    tmp += (numpy.roll(by, -1, axis=2) - numpy.roll(by,  1, axis=2))

                for p in range(self.chunks[chunk_number]['dblocks']):
                    tmp[p,...] /= self.cell_size[self.chunks[chunk_number]['levels'][p]]
                dset -= 0.5 * self.attributes['resistivity'] * tmp
        elif dataset == 'elec':
            b1 = self.__read_binary_data__('magy', chunk_number)
            b2 = self.__read_binary_data__('magz', chunk_number)
            v1 = self.__read_binary_data__('vely', chunk_number)
            v2 = self.__read_binary_data__('velz', chunk_number)
            dtmp = v1 * b2 - v2 * b1
            if self.attributes['resistivity'] > 0:
                if self.attributes['ndims'] == 3:
                    tmp  = (numpy.roll(b1,  1, axis=1) - numpy.roll(b1, -1, axis=1))
                    tmp += (numpy.roll(b2, -1, axis=2) - numpy.roll(b2,  1, axis=2))
                else:
                    tmp  = (numpy.roll(b2, -1, axis=1) - numpy.roll(b2,  1, axis=1))

                for p in range(self.chunks[chunk_number]['dblocks']):
                    tmp[p,...] /= self.cell_size[self.chunks[chunk_number]['levels'][p]]
                dtmp -= 0.5 * self.attributes['resistivity'] * tmp
            dset  = dtmp**2

            b1 = self.__read_binary_data__('magx', chunk_number)
            v1 = self.__read_binary_data__('velx', chunk_number)
            dtmp = v2 * b1 - v1 * b2
            if self.attributes['resistivity'] > 0:
                if self.attributes['ndims'] == 3:
                    tmp  = (numpy.roll(b1, -1, axis=1) - numpy.roll(b1,  1, axis=1))
                    tmp += (numpy.roll(b2,  1, axis=3) - numpy.roll(b2, -1, axis=3))
                else:
                    tmp  = (numpy.roll(b2,  1, axis=2) - numpy.roll(b2, -1, axis=2))

                for p in range(self.chunks[chunk_number]['dblocks']):
                    tmp[p,...] /= self.cell_size[self.chunks[chunk_number]['levels'][p]]
                dtmp -= 0.5 * self.attributes['resistivity'] * tmp
            dset += dtmp**2

            b2 = self.__read_binary_data__('magy', chunk_number)
            v2 = self.__read_binary_data__('vely', chunk_number)
            dtmp = v1 * b2 - v2 * b1
            if self.attributes['resistivity'] > 0:
                if self.attributes['ndims'] == 3:
                    tmp  = (numpy.roll(b1,  1, axis=2) - numpy.roll(b1, -1, axis=2))
                    tmp += (numpy.roll(b2, -1, axis=3) - numpy.roll(b2,  1, axis=3))
                else:
                    tmp  = (numpy.roll(b1,  1, axis=1) - numpy.roll(b1, -1, axis=1))
                    tmp += (numpy.roll(b2, -1, axis=2) - numpy.roll(b2,  1, axis=2))

                for p in range(self.chunks[chunk_number]['dblocks']):
                    tmp[p,...] /= self.cell_size[self.chunks[chunk_number]['levels'][p]]
                dtmp -= 0.5 * self.attributes['resistivity'] * tmp
            dset += dtmp**2
            dset = numpy.sqrt(dset)
        elif dataset == 'evec':
            b1 = self.__read_binary_data__('magy', chunk_number)
            b2 = self.__read_binary_data__('magz', chunk_number)
            v1 = self.__read_binary_data__('vely', chunk_number)
            v2 = self.__read_binary_data__('velz', chunk_number)
            wx = v1 * b2 - v2 * b1
            if self.attributes['resistivity'] > 0:
                if self.attributes['ndims'] == 3:
                    tmp  = (numpy.roll(b1,  1, axis=1) - numpy.roll(b1, -1, axis=1))
                    tmp += (numpy.roll(b2, -1, axis=2) - numpy.roll(b2,  1, axis=2))
                else:
                    tmp  = (numpy.roll(b2, -1, axis=1) - numpy.roll(b2,  1, axis=1))

                for p in range(self.chunks[chunk_number]['dblocks']):
                    tmp[p,...] /= self.cell_size[self.chunks[chunk_number]['levels'][p]]
                wx -= 0.5 * self.attributes['resistivity'] * tmp

            b1 = self.__read_binary_data__('magx', chunk_number)
            v1 = self.__read_binary_data__('velx', chunk_number)
            wy = v2 * b1 - v1 * b2
            if self.attributes['resistivity'] > 0:
                if self.attributes['ndims'] == 3:
                    tmp  = (numpy.roll(b1, -1, axis=1) - numpy.roll(b1,  1, axis=1))
                    tmp += (numpy.roll(b2,  1, axis=3) - numpy.roll(b2, -1, axis=3))
                else:
                    tmp  = (numpy.roll(b2,  1, axis=2) - numpy.roll(b2, -1, axis=2))

                for p in range(self.chunks[chunk_number]['dblocks']):
                    tmp[p,...] /= self.cell_size[self.chunks[chunk_number]['levels'][p]]
                wy -= 0.5 * self.attributes['resistivity'] * tmp

            b2 = self.__read_binary_data__('magy', chunk_number)
            v2 = self.__read_binary_data__('vely', chunk_number)
            wz = v1 * b2 - v2 * b1
            if self.attributes['resistivity'] > 0:
                if self.attributes['ndims'] == 3:
                    tmp  = (numpy.roll(b1,  1, axis=2) - numpy.roll(b1, -1, axis=2))
                    tmp += (numpy.roll(b2, -1, axis=3) - numpy.roll(b2,  1, axis=3))
                else:
                    tmp  = (numpy.roll(b1,  1, axis=1) - numpy.roll(b1, -1, axis=1))
                    tmp += (numpy.roll(b2, -1, axis=2) - numpy.roll(b2,  1, axis=2))

                for p in range(self.chunks[chunk_number]['dblocks']):
                    tmp[p,...] /= self.cell_size[self.chunks[chunk_number]['levels'][p]]
                wz -= 0.5 * self.attributes['resistivity'] * tmp

            dset = [wx, wy, wz]
        elif dataset == 'eint':
            dset = self.__read_binary_data__('pres', chunk_number)
            dset *= 1.0 / (self.attributes('adiabatic_index') - 1)
        elif dataset == 'temp':
            dset = self.__read_binary_data__('pres', chunk_number) / self.__read_binary_data__('dens', chunk_number)
        elif dataset == 'etot':
            tmp = self.__read_binary_data__('velx', chunk_number)
            dset = tmp**2
            tmp = self.__read_binary_data__('vely', chunk_number)
            dset += tmp**2
            tmp = self.__read_binary_data__('velz', chunk_number)
            dset += tmp**2
            tmp = self.__read_binary_data__('dens', chunk_number)
            dset *= tmp
            tmp = self.__read_binary_data__('pres', chunk_number)
            dset += 2.0 / (self.attributes('adiabatic_index') - 1) * tmp
            if 'magn' in self.variables:
                tmp = self.__read_binary_data__('magx', chunk_number)
                dset = tmp**2
                tmp = self.__read_binary_data__('magy', chunk_number)
                dset += tmp**2
                tmp = self.__read_binary_data__('magz', chunk_number)
                dset += tmp**2
            dset *= 0.5
        elif dataset == 'lore':
            tmp = self.__read_binary_data__('velx', chunk_number)
            dset = tmp**2
            tmp = self.__read_binary_data__('vely', chunk_number)
            dset += tmp**2
            tmp = self.__read_binary_data__('velz', chunk_number)
            dset += tmp**2
            dset = 1.0 / numpy.sqrt(1.0 - dset)
        elif dataset == 'divv':
            p = self.attributes['ndims']
            tmp = self.__read_binary_data__('velx', chunk_number)
            dset  = (numpy.roll(tmp, -1, axis=p) \
                   - numpy.roll(tmp,  1, axis=p))
            p -= 1
            tmp = self.__read_binary_data__('vely', chunk_number)
            dset += (numpy.roll(tmp, -1, axis=p) \
                   - numpy.roll(tmp,  1, axis=p))
            p -= 1
            if p >= 0:
                tmp = self.__read_binary_data__('velz', chunk_number)
                dset += (numpy.roll(tmp, -1, axis=p) \
                       - numpy.roll(tmp,  0, axis=p))

            dset *= 0.5
            for p in range(self.chunks[chunk_number]['dblocks']):
                dset[p,...] /= self.cell_size[self.chunks[chunk_number]['levels'][p]]
        elif dataset == 'vorx':
            if self.attributes['ndims'] == 3:
                tmp = self.__read_binary_data__('vely', chunk_number)
                dset  = (numpy.roll(tmp,  1, axis=1) \
                       - numpy.roll(tmp, -1, axis=1))
                tmp = self.__read_binary_data__('velz', chunk_number)
                dset += (numpy.roll(tmp, -1, axis=2) \
                       - numpy.roll(tmp,  1, axis=2))
            else:
                tmp = self.__read_binary_data__('velz', chunk_number)
                dset  = (numpy.roll(tmp, -1, axis=1) \
                       - numpy.roll(tmp,  1, axis=1))

            dset *= 0.5
            for p in range(self.chunks[chunk_number]['dblocks']):
                dset[p,...] /= self.cell_size[self.chunks[chunk_number]['levels'][p]]
        elif dataset == 'vory':
            if self.attributes['ndims'] == 3:
                tmp = self.__read_binary_data__('velx', chunk_number)
                dset  = (numpy.roll(tmp, -1, axis=1) \
                       - numpy.roll(tmp,  1, axis=1))
                tmp = self.__read_binary_data__('velz', chunk_number)
                dset += (numpy.roll(tmp,  1, axis=3) \
                       - numpy.roll(tmp, -1, axis=3))
            else:
                tmp = self.__read_binary_data__('velz', chunk_number)
                dset  = (numpy.roll(tmp,  1, axis=2) \
                       - numpy.roll(tmp, -1, axis=2))

            dset *= 0.5
            for p in range(self.chunks[chunk_number]['dblocks']):
                dset[p,...] /= self.cell_size[self.chunks[chunk_number]['levels'][p]]
        elif dataset == 'vorz':
            if self.attributes['ndims'] == 3:
                tmp = self.__read_binary_data__('velx', chunk_number)
                dset  = (numpy.roll(tmp,  1, axis=2) \
                       - numpy.roll(tmp, -1, axis=2))
                tmp = self.__read_binary_data__('vely', chunk_number)
                dset += (numpy.roll(tmp, -1, axis=3) \
                       - numpy.roll(tmp,  1, axis=3))
            else:
                tmp = self.__read_binary_data__('velx', chunk_number)
                dset  = (numpy.roll(tmp,  1, axis=1) \
                       - numpy.roll(tmp, -1, axis=1))
                tmp = self.__read_binary_data__('vely', chunk_number)
                dset += (numpy.roll(tmp, -1, axis=2) \
                       - numpy.roll(tmp,  1, axis=2))

            dset *= 0.5
            for p in range(self.chunks[chunk_number]['dblocks']):
                dset[p,...] /= self.cell_size[self.chunks[chunk_number]['levels'][p]]
        elif dataset == 'wvec':
            if self.attributes['ndims'] == 3:
                tmp = self.__read_binary_data__('velx', chunk_number)
                wy  = (numpy.roll(tmp, -1, axis=1) \
                     - numpy.roll(tmp,  1, axis=1))
                wz  = (numpy.roll(tmp,  1, axis=2) \
                     - numpy.roll(tmp, -1, axis=2))
                tmp = self.__read_binary_data__('vely', chunk_number)
                wz += (numpy.roll(tmp, -1, axis=3) \
                     - numpy.roll(tmp,  1, axis=3))
                wx  = (numpy.roll(tmp,  1, axis=1) \
                     - numpy.roll(tmp, -1, axis=1))
                tmp = self.__read_binary_data__('velz', chunk_number)
                wx += (numpy.roll(tmp, -1, axis=2) \
                     - numpy.roll(tmp,  1, axis=2))
                wy += (numpy.roll(tmp,  1, axis=3) \
                     - numpy.roll(tmp, -1, axis=3))
            else:
                tmp = self.__read_binary_data__('velx', chunk_number)
                wz  = (numpy.roll(tmp,  1, axis=1) \
                     - numpy.roll(tmp, -1, axis=1))
                tmp = self.__read_binary_data__('vely', chunk_number)
                wz += (numpy.roll(tmp, -1, axis=2) \
                     - numpy.roll(tmp,  1, axis=2))
                tmp = self.__read_binary_data__('velz', chunk_number)
                wx  = (numpy.roll(tmp, -1, axis=1) \
                     - numpy.roll(tmp,  1, axis=1))
                wy  = (numpy.roll(tmp,  1, axis=2) \
                     - numpy.roll(tmp, -1, axis=2))

            for p in range(self.chunks[chunk_number]['dblocks']):
                h = 2 * self.cell_size[self.chunks[chunk_number]['levels'][p]]
                wx[p,...] /= h
                wy[p,...] /= h
                wz[p,...] /= h

            dset = [wx, wy, wz]
        elif dataset == 'vort':
            if self.attributes['ndims'] == 3:
                tmp = self.__read_binary_data__('velx', chunk_number)
                wy  = (numpy.roll(tmp, -1, axis=1) \
                     - numpy.roll(tmp,  1, axis=1))
                wz  = (numpy.roll(tmp,  1, axis=2) \
                     - numpy.roll(tmp, -1, axis=2))
                tmp = self.__read_binary_data__('vely', chunk_number)
                wz += (numpy.roll(tmp, -1, axis=3) \
                     - numpy.roll(tmp,  1, axis=3))
                wx  = (numpy.roll(tmp,  1, axis=1) \
                     - numpy.roll(tmp, -1, axis=1))
                tmp = self.__read_binary_data__('velz', chunk_number)
                wx += (numpy.roll(tmp, -1, axis=2) \
                     - numpy.roll(tmp,  1, axis=2))
                wy += (numpy.roll(tmp,  1, axis=3) \
                     - numpy.roll(tmp, -1, axis=3))
            else:
                tmp = self.__read_binary_data__('velx', chunk_number)
                wz  = (numpy.roll(tmp,  1, axis=1) \
                     - numpy.roll(tmp, -1, axis=1))
                tmp = self.__read_binary_data__('vely', chunk_number)
                wz += (numpy.roll(tmp, -1, axis=2) \
                     - numpy.roll(tmp,  1, axis=2))
                tmp = self.__read_binary_data__('velz', chunk_number)
                wx  = (numpy.roll(tmp, -1, axis=1) \
                     - numpy.roll(tmp,  1, axis=1))
                wy  = (numpy.roll(tmp,  1, axis=2) \
                     - numpy.roll(tmp, -1, axis=2))

            dset = 0.5 * numpy.sqrt(wx**2 + wy**2 + wz**2)
            for p in range(self.chunks[chunk_number]['dblocks']):
                dset[p,...] /= self.cell_size[self.chunks[chunk_number]['levels'][p]]
        elif dataset == 'dvxdx':
            p = self.attributes['ndims']
            tmp = self.__read_binary_data__('velx', chunk_number)
            dset = numpy.roll(tmp, -1, axis=p) - numpy.roll(tmp,  1, axis=p)
            dset *= 0.5
            for p in range(self.chunks[chunk_number]['dblocks']):
                dset[p,...] /= self.cell_size[self.chunks[chunk_number]['levels'][p]]
        elif dataset == 'dvxdy':
            p = self.attributes['ndims'] - 1
            tmp = self.__read_binary_data__('velx', chunk_number)
            dset = numpy.roll(tmp, -1, axis=p) - numpy.roll(tmp,  1, axis=p)
            dset *= 0.5
            for p in range(self.chunks[chunk_number]['dblocks']):
                dset[p,...] /= self.cell_size[self.chunks[chunk_number]['levels'][p]]
        elif dataset == 'dvxdz':
            p = self.attributes['ndims'] - 2
            tmp = self.__read_binary_data__('velx', chunk_number)
            if p >= 0:
                dset = numpy.roll(tmp, -1, axis=p) - numpy.roll(tmp,  1, axis=p)
                dset *= 0.5
                for p in range(self.chunks[chunk_number]['dblocks']):
                    dset[p,...] /= self.cell_size[self.chunks[chunk_number]['levels'][p]]
            else:
                dset = numpy.zeros_like(tmp)
        elif dataset == 'dvydx':
            p = self.attributes['ndims']
            tmp = self.__read_binary_data__('vely', chunk_number)
            dset = numpy.roll(tmp, -1, axis=p) - numpy.roll(tmp,  1, axis=p)
            dset *= 0.5
            for p in range(self.chunks[chunk_number]['dblocks']):
                dset[p,...] /= self.cell_size[self.chunks[chunk_number]['levels'][p]]
        elif dataset == 'dvydy':
            p = self.attributes['ndims'] - 1
            tmp = self.__read_binary_data__('vely', chunk_number)
            dset = numpy.roll(tmp, -1, axis=p) - numpy.roll(tmp,  1, axis=p)
            dset *= 0.5
            for p in range(self.chunks[chunk_number]['dblocks']):
                dset[p,...] /= self.cell_size[self.chunks[chunk_number]['levels'][p]]
        elif dataset == 'dvydz':
            p = self.attributes['ndims'] - 2
            tmp = self.__read_binary_data__('vely', chunk_number)
            if p >= 0:
                dset = numpy.roll(tmp, -1, axis=p) - numpy.roll(tmp,  1, axis=p)
                dset *= 0.5
                for p in range(self.chunks[chunk_number]['dblocks']):
                    dset[p,...] /= self.cell_size[self.chunks[chunk_number]['levels'][p]]
            else:
                dset = numpy.zeros_like(tmp)
        elif dataset == 'dvzdx':
            p = self.attributes['ndims']
            tmp = self.__read_binary_data__('velz', chunk_number)
            dset = numpy.roll(tmp, -1, axis=p) - numpy.roll(tmp,  1, axis=p)
            dset *= 0.5
            for p in range(self.chunks[chunk_number]['dblocks']):
                dset[p,...] /= self.cell_size[self.chunks[chunk_number]['levels'][p]]
        elif dataset == 'dvzdy':
            p = self.attributes['ndims'] - 1
            tmp = self.__read_binary_data__('velz', chunk_number)
            dset = numpy.roll(tmp, -1, axis=p) - numpy.roll(tmp,  1, axis=p)
            dset *= 0.5
            for p in range(self.chunks[chunk_number]['dblocks']):
                dset[p,...] /= self.cell_size[self.chunks[chunk_number]['levels'][p]]
        elif dataset == 'dvzdz':
            p = self.attributes['ndims'] - 2
            tmp = self.__read_binary_data__('velz', chunk_number)
            if p >= 0:
                dset = numpy.roll(tmp, -1, axis=p) - numpy.roll(tmp,  1, axis=p)
                dset *= 0.5
                for p in range(self.chunks[chunk_number]['dblocks']):
                    dset[p,...] /= self.cell_size[self.chunks[chunk_number]['levels'][p]]
            else:
                dset = numpy.zeros_like(tmp)
        elif dataset == 'divb':
            p = self.attributes['ndims']
            tmp = self.__read_binary_data__('magx', chunk_number)
            dset  = (numpy.roll(tmp, -1, axis=p) \
                   - numpy.roll(tmp,  1, axis=p))
            p -= 1
            tmp = self.__read_binary_data__('magy', chunk_number)
            dset += (numpy.roll(tmp, -1, axis=p) \
                   - numpy.roll(tmp,  1, axis=p))
            p -= 1
            if p >= 0:
                tmp = self.__read_binary_data__('magz', chunk_number)
                dset += (numpy.roll(tmp, -1, axis=p) \
                       - numpy.roll(tmp,  0, axis=p))

            dset *= 0.5
            for p in range(self.chunks[chunk_number]['dblocks']):
                dset[p,...] /= self.cell_size[self.chunks[chunk_number]['levels'][p]]
        elif dataset == 'curx':
            if self.attributes['ndims'] == 3:
                tmp = self.__read_binary_data__('magy', chunk_number)
                dset  = (numpy.roll(tmp,  1, axis=1) \
                       - numpy.roll(tmp, -1, axis=1))
                tmp = self.__read_binary_data__('magz', chunk_number)
                dset += (numpy.roll(tmp, -1, axis=2) \
                       - numpy.roll(tmp,  1, axis=2))
            else:
                tmp = self.__read_binary_data__('magz', chunk_number)
                dset  = (numpy.roll(tmp, -1, axis=1) \
                       - numpy.roll(tmp,  1, axis=1))

            dset *= 0.5
            for p in range(self.chunks[chunk_number]['dblocks']):
                dset[p,...] /= self.cell_size[self.chunks[chunk_number]['levels'][p]]
        elif dataset == 'cury':
            if self.attributes['ndims'] == 3:
                tmp = self.__read_binary_data__('magx', chunk_number)
                dset  = (numpy.roll(tmp, -1, axis=1) \
                       - numpy.roll(tmp,  1, axis=1))
                tmp = self.__read_binary_data__('magz', chunk_number)
                dset += (numpy.roll(tmp,  1, axis=3) \
                       - numpy.roll(tmp, -1, axis=3))
            else:
                tmp = self.__read_binary_data__('magz', chunk_number)
                dset  = (numpy.roll(tmp,  1, axis=2) \
                       - numpy.roll(tmp, -1, axis=2))

            dset *= 0.5
            for p in range(self.chunks[chunk_number]['dblocks']):
                dset[p,...] /= self.cell_size[self.chunks[chunk_number]['levels'][p]]
        elif dataset == 'curz':
            if self.attributes['ndims'] == 3:
                tmp = self.__read_binary_data__('magx', chunk_number)
                dset  = (numpy.roll(tmp,  1, axis=2) \
                       - numpy.roll(tmp, -1, axis=2))
                tmp = self.__read_binary_data__('magy', chunk_number)
                dset += (numpy.roll(tmp, -1, axis=3) \
                       - numpy.roll(tmp,  1, axis=3))
            else:
                tmp = self.__read_binary_data__('magx', chunk_number)
                dset  = (numpy.roll(tmp,  1, axis=1) \
                       - numpy.roll(tmp, -1, axis=1))
                tmp = self.__read_binary_data__('magy', chunk_number)
                dset += (numpy.roll(tmp, -1, axis=2) \
                       - numpy.roll(tmp,  1, axis=2))

            dset *= 0.5
            for p in range(self.chunks[chunk_number]['dblocks']):
                dset[p,...] /= self.cell_size[self.chunks[chunk_number]['levels'][p]]
        elif dataset == 'jvec':
            if self.attributes['ndims'] == 3:
                tmp = self.__read_binary_data__('magx', chunk_number)
                wy  = (numpy.roll(tmp, -1, axis=1) \
                     - numpy.roll(tmp,  1, axis=1))
                wz  = (numpy.roll(tmp,  1, axis=2) \
                     - numpy.roll(tmp, -1, axis=2))
                tmp = self.__read_binary_data__('magy', chunk_number)
                wz += (numpy.roll(tmp, -1, axis=3) \
                     - numpy.roll(tmp,  1, axis=3))
                wx  = (numpy.roll(tmp,  1, axis=1) \
                     - numpy.roll(tmp, -1, axis=1))
                tmp = self.__read_binary_data__('magz', chunk_number)
                wx += (numpy.roll(tmp, -1, axis=2) \
                     - numpy.roll(tmp,  1, axis=2))
                wy += (numpy.roll(tmp,  1, axis=3) \
                     - numpy.roll(tmp, -1, axis=3))
            else:
                tmp = self.__read_binary_data__('magx', chunk_number)
                wz  = (numpy.roll(tmp,  1, axis=1) \
                     - numpy.roll(tmp, -1, axis=1))
                tmp = self.__read_binary_data__('magy', chunk_number)
                wz += (numpy.roll(tmp, -1, axis=2) \
                     - numpy.roll(tmp,  1, axis=2))
                tmp = self.__read_binary_data__('magz', chunk_number)
                wx  = (numpy.roll(tmp, -1, axis=1) \
                     - numpy.roll(tmp,  1, axis=1))
                wy  = (numpy.roll(tmp,  1, axis=2) \
                     - numpy.roll(tmp, -1, axis=2))

            for p in range(self.chunks[chunk_number]['dblocks']):
                h = 2 * self.cell_size[self.chunks[chunk_number]['levels'][p]]
                wx[p,...] /= h
                wy[p,...] /= h
                wz[p,...] /= h

            dset = [wx, wy, wz]
        elif dataset == 'curr':
            if self.attributes['ndims'] == 3:
                tmp = self.__read_binary_data__('magx', chunk_number)
                wy  = (numpy.roll(tmp, -1, axis=1) \
                     - numpy.roll(tmp,  1, axis=1))
                wz  = (numpy.roll(tmp,  1, axis=2) \
                     - numpy.roll(tmp, -1, axis=2))
                tmp = self.__read_binary_data__('magy', chunk_number)
                wz += (numpy.roll(tmp, -1, axis=3) \
                     - numpy.roll(tmp,  1, axis=3))
                wx  = (numpy.roll(tmp,  1, axis=1) \
                     - numpy.roll(tmp, -1, axis=1))
                tmp = self.__read_binary_data__('magz', chunk_number)
                wx += (numpy.roll(tmp, -1, axis=2) \
                     - numpy.roll(tmp,  1, axis=2))
                wy += (numpy.roll(tmp,  1, axis=3) \
                     - numpy.roll(tmp, -1, axis=3))
            else:
                tmp = self.__read_binary_data__('magx', chunk_number)
                wz  = (numpy.roll(tmp,  1, axis=1) \
                     - numpy.roll(tmp, -1, axis=1))
                tmp = self.__read_binary_data__('magy', chunk_number)
                wz += (numpy.roll(tmp, -1, axis=2) \
                     - numpy.roll(tmp,  1, axis=2))
                tmp = self.__read_binary_data__('magz', chunk_number)
                wx  = (numpy.roll(tmp, -1, axis=1) \
                     - numpy.roll(tmp,  1, axis=1))
                wy  = (numpy.roll(tmp,  1, axis=2) \
                     - numpy.roll(tmp, -1, axis=2))

            dset = 0.5 * numpy.sqrt(wx**2 + wy**2 + wz**2)
            for p in range(self.chunks[chunk_number]['dblocks']):
                dset[p,...] /= self.cell_size[self.chunks[chunk_number]['levels'][p]]
        elif dataset == 'dbxdx':
            p = self.attributes['ndims']
            tmp = self.__read_binary_data__('magx', chunk_number)
            dset = numpy.roll(tmp, -1, axis=p) - numpy.roll(tmp,  1, axis=p)
            dset *= 0.5
            for p in range(self.chunks[chunk_number]['dblocks']):
                dset[p,...] /= self.cell_size[self.chunks[chunk_number]['levels'][p]]
        elif dataset == 'dbxdy':
            p = self.attributes['ndims'] - 1
            tmp = self.__read_binary_data__('magx', chunk_number)
            dset = numpy.roll(tmp, -1, axis=p) - numpy.roll(tmp,  1, axis=p)
            dset *= 0.5
            for p in range(self.chunks[chunk_number]['dblocks']):
                dset[p,...] /= self.cell_size[self.chunks[chunk_number]['levels'][p]]
        elif dataset == 'dbxdz':
            p = self.attributes['ndims'] - 2
            tmp = self.__read_binary_data__('magx', chunk_number)
            if p >= 0:
                dset = numpy.roll(tmp, -1, axis=p) - numpy.roll(tmp,  1, axis=p)
                dset *= 0.5
                for p in range(self.chunks[chunk_number]['dblocks']):
                    dset[p,...] /= self.cell_size[self.chunks[chunk_number]['levels'][p]]
            else:
                dset = numpy.zeros_like(tmp)
        elif dataset == 'dbydx':
            p = self.attributes['ndims']
            tmp = self.__read_binary_data__('magy', chunk_number)
            dset = numpy.roll(tmp, -1, axis=p) - numpy.roll(tmp,  1, axis=p)
            dset *= 0.5
            for p in range(self.chunks[chunk_number]['dblocks']):
                dset[p,...] /= self.cell_size[self.chunks[chunk_number]['levels'][p]]
        elif dataset == 'dbydy':
            p = self.attributes['ndims'] - 1
            tmp = self.__read_binary_data__('magy', chunk_number)
            dset = numpy.roll(tmp, -1, axis=p) - numpy.roll(tmp,  1, axis=p)
            dset *= 0.5
            for p in range(self.chunks[chunk_number]['dblocks']):
                dset[p,...] /= self.cell_size[self.chunks[chunk_number]['levels'][p]]
        elif dataset == 'dbydz':
            p = self.attributes['ndims'] - 2
            tmp = self.__read_binary_data__('magy', chunk_number)
            if p >= 0:
                dset = numpy.roll(tmp, -1, axis=p) - numpy.roll(tmp,  1, axis=p)
                dset *= 0.5
                for p in range(self.chunks[chunk_number]['dblocks']):
                    dset[p,...] /= self.cell_size[self.chunks[chunk_number]['levels'][p]]
            else:
                dset = numpy.zeros_like(tmp)
        elif dataset == 'dbzdx':
            p = self.attributes['ndims']
            tmp = self.__read_binary_data__('magz', chunk_number)
            dset = numpy.roll(tmp, -1, axis=p) - numpy.roll(tmp,  1, axis=p)
            dset *= 0.5
            for p in range(self.chunks[chunk_number]['dblocks']):
                dset[p,...] /= self.cell_size[self.chunks[chunk_number]['levels'][p]]
        elif dataset == 'dbzdy':
            p = self.attributes['ndims'] - 1
            tmp = self.__read_binary_data__('magz', chunk_number)
            dset = numpy.roll(tmp, -1, axis=p) - numpy.roll(tmp,  1, axis=p)
            dset *= 0.5
            for p in range(self.chunks[chunk_number]['dblocks']):
                dset[p,...] /= self.cell_size[self.chunks[chunk_number]['levels'][p]]
        elif dataset == 'dbzdz':
            p = self.attributes['ndims'] - 2
            tmp = self.__read_binary_data__('magz', chunk_number)
            if p >= 0:
                dset = numpy.roll(tmp, -1, axis=p) - numpy.roll(tmp,  1, axis=p)
                dset *= 0.5
                for p in range(self.chunks[chunk_number]['dblocks']):
                    dset[p,...] /= self.cell_size[self.chunks[chunk_number]['levels'][p]]
            else:
                dset = numpy.zeros_like(tmp)
        else:
            dset = self.__read_binary_data__(dataset, chunk_number)

        return dset


    def dataset(self, dataset_name, extent=None, maxlev=None, shrink=1, \
                           interpolation='rebin', order=3, progress=False):
        """
            Function returns dataset of the requested variable resampled to
            the uniform mesh.
        """
        from .interpolation import interpolate
        import numpy, sys

        if self.dataformat == None:
            raise Exception("Snapshot object has not been properly initialized!")

        if not dataset_name in self.variables:
            raise Exception("Dataset '{}' is not available!\nAvailable datasets: {}\n".format(dataset_name, list(self.variables.keys())))

        dlo = numpy.array([self.attributes['xmin'], self.attributes['ymin']])
        dup = numpy.array([self.attributes['xmax'], self.attributes['ymax']])
        dln = numpy.array([self.attributes['xlen'], self.attributes['ylen']])
        if self.attributes['ndims'] == 3:
            dlo = numpy.append(dlo, self.attributes['zmin'])
            dup = numpy.append(dup, self.attributes['zmax'])
            dln = numpy.append(dln, self.attributes['zlen'])

        slo = numpy.array(dlo)
        sup = numpy.array(dup)
        if extent != None:
            if len(extent) != 2 * self.attributes['ndims']:
                raise Exception("Wrong dimensions of the argument 'extent'!")
            slo = numpy.array(extent)[0::2]
            sup = numpy.array(extent)[1::2]
            if any(slo > dup) or any(sup < dlo) or any (slo >= sup):
                raise Exception("Wrong order of the dimensions in the argument 'extent'!")

        if maxlev != None:
            if isinstance(maxlev, (int, numpy.int32)):
                if 1 <= maxlev <= self.attributes['maxlev']:
                    shrink = 2**(self.attributes['maxlev']-maxlev)
                elif maxlev > self.attributes['maxlev']:
                    raise Exception("Argument 'maxlev' should be between 1 and {}.\n".format(self.attributes['maxlev']))
            else:
                raise Exception("Argument 'maxlev' must be an integer between 1 and {}.\n".format(self.attributes['maxlev']))

        bm = numpy.array([ self.attributes['ncells'] ]*self.attributes['ndims'])

        if self.attributes['ndims'] == 3:
            rm = numpy.array([self.attributes['zblocks'], self.attributes['yblocks'], self.attributes['xblocks']])
        else:
            rm = numpy.array([self.attributes['yblocks'], self.attributes['xblocks']])

        dm = rm * self.attributes['ncells'] * 2**(self.attributes['toplev'] - 1) // shrink
        ll = numpy.array(numpy.floor(dm[::-1] * (slo - dlo) / dln), dtype='int')
        uu = numpy.array( numpy.ceil(dm[::-1] * (sup - dlo) / dln), dtype='int')
        dm  = (uu - ll)[::-1]
        if dataset_name in [ 'velocity', 'vorticity', 'magnetic field', 'current density', 'electric field']:
            arr = [ numpy.zeros(dm[:]), numpy.zeros(dm[:]), numpy.zeros(dm[:]) ]
        else:
            arr = numpy.zeros(dm[:])

        if progress:
            sys.stdout.write("Snapshot's path:\n  '{}'\n".format(self.path))
            m = 0

        for n in range(self.attributes['nchunks']):

            if self.chunks[n]['dblocks'] > 0:

                dset = self.__get_dataset__(dataset_name, n)

                for p in range(self.chunks[n]['dblocks']):
                    nl = 2**(self.attributes['toplev'] - self.chunks[n]['levels'][p])
                    if nl <= shrink:
                        method = 'rebin'
                    else:
                        method = interpolation
                    cm = bm[0:self.attributes['ndims']] * nl // shrink
                    il = self.chunks[n]['coords'][p,:][0:self.attributes['ndims']] * cm[0:self.attributes['ndims']]
                    iu = il + cm[0:self.attributes['ndims']]

                    if all(iu[:] > ll[:]) and all(il[:] < uu[:]):
                        nb = il[:] - ll[:]
                        ne = iu[:] - ll[:]
                        ib = numpy.maximum(nb[:], 0)
                        ie = numpy.minimum(ne[:], uu[:] - ll[:])
                        jb = ib[:] - nb[:]
                        je = ie[:] - ne[:] + cm[:]

                        if isinstance(dset, (list, tuple)):
                            for di in range(3):
                                if self.attributes['ndims'] == 3:
                                    arr[di][ib[2]:ie[2],ib[1]:ie[1],ib[0]:ie[0]] = interpolate(dset[di][p,:,:,:], cm, self.attributes['nghosts'], method=method, order=order)[jb[2]:je[2],jb[1]:je[1],jb[0]:je[0]]
                                else:
                                    arr[di][ib[1]:ie[1],ib[0]:ie[0]] = interpolate(dset[di][p,:,:], cm, self.attributes['nghosts'], method=method, order=order)[jb[1]:je[1],jb[0]:je[0]]
                        else:
                            if self.attributes['ndims'] == 3:
                                arr[ib[2]:ie[2],ib[1]:ie[1],ib[0]:ie[0]] = interpolate(dset[p,:,:,:], cm, self.attributes['nghosts'], method=method, order=order)[jb[2]:je[2],jb[1]:je[1],jb[0]:je[0]]
                            else:
                                arr[ib[1]:ie[1],ib[0]:ie[0]] = interpolate(dset[p,:,:], cm, self.attributes['nghosts'], method=method, order=order)[jb[1]:je[1],jb[0]:je[0]]

                    if progress:
                        cfile = self.chunkname.format(n)
                        m += 1
                        sys.stdout.write('\r')
                        sys.stdout.write("Reading '{}' from '{}': block {} of {}".format(dataset_name, cfile, m, self.attributes['nleafs']))
                        sys.stdout.flush()

        if (progress):
            sys.stdout.write('\n')
            sys.stdout.flush()

        return arr


    def dataset_to_vtk(self, dataset_name, label=None, compression=None, compression_level=19, progress=False):
        """
            Function converts dataset of the requested variable to AMR VTK file.
        """
        import numpy, os, sys
        from .octree import OcBase, OcNode
        from .vtkio import WriteVTK

        if self.dataformat == None:
            raise Exception("Snapshot object has not been properly initialized!")

        if not dataset_name in self.variables:
            raise Exception("Dataset '{}' is not available!\nAvailable datasets: {}\n".format(dataset_name, list(self.variables.keys())))

        if label == None:
            label = dataset_name

        if self.attributes['ndims'] < 3:
            raise Exception("Conversion to OverlappedAMR works only with 3D snapshots!")

        base = OcBase([self.attributes['xmin'], self.attributes['ymin'], self.attributes['zmin']], \
                      [self.attributes['xlen'], self.attributes['ylen'], self.attributes['zlen']], \
                      [self.attributes['xblocks'], self.attributes['yblocks'], self.attributes['zblocks']])

        ng = self.attributes['nghosts']

        if progress:
            sys.stdout.write("Snapshot's path:\n  '{}'\n".format(self.path))
            m = 0

        for n in range(self.attributes['nchunks']):

            if self.chunks[n]['dblocks'] > 0:

                dset = self.__get_dataset__(dataset_name, n)

                for p in range(self.chunks[n]['dblocks']):

                    lv = self.chunks[n]['levels'][p] - 1

                    center = (self.chunks[n]['bounds'][p,0,:] + self.chunks[n]['bounds'][p,1,:]) / 2
                    base.createNodeBranch(center, lv)
                    if isinstance(dset, (list, tuple)):
                        base.setNodeData(center, lv, (dset[0][p,ng:-ng,ng:-ng,ng:-ng], dset[1][p,ng:-ng,ng:-ng,ng:-ng], dset[2][p,ng:-ng,ng:-ng,ng:-ng]))
                    else:
                        base.setNodeData(center, lv, dset[p,ng:-ng,ng:-ng,ng:-ng])

                    if progress:
                        cfile = self.chunkname.format(n)
                        m += 1
                        sys.stdout.write('\r')
                        sys.stdout.write("Reading '{}' from '{}': block {} of {}".format(dataset_name, cfile, m, self.attributes['nleafs']))
                        sys.stdout.flush()

        if (progress):
            sys.stdout.write('\n')
            sys.stdout.flush()

        if progress:
            sys.stdout.write("Populating AMR structure\n")
        base.populateNodeData()

        if progress:
            sys.stdout.write("Generating OverlappingAMR VTK files\n")

        bm = numpy.array([ self.attributes['ncells'] ]*self.attributes['ndims'])

        if self.attributes['ndims'] == 3:
            rm = numpy.array([self.attributes['zblocks'], self.attributes['yblocks'], self.attributes['xblocks']])
        else:
            rm = numpy.array([self.attributes['yblocks'], self.attributes['xblocks']])

        ofile = "{}_{:06d}.vthb".format(dataset_name, self.attributes['isnap'])
        opath = "{}_{:06d}".format(dataset_name, self.attributes['isnap'])
        if not os.path.exists(opath):
            os.makedirs(opath)
        with open(ofile, 'w') as vtk:
            vtk.write('<VTKFile type="vtkOverlappingAMR" version="1.1" ' + \
                        'byte_order="LittleEndian" header_type="UInt64">\n')
            vtk.write('  <vtkOverlappingAMR ' + \
                        'origin="{} {} {}" '.format(*base.lower) + \
                        'grid_description="XYZ">\n')

            fmt = '{}_{:0' + str(len(str(self.attributes['toplev']))) + '}_{:0' + str(len(str(base.nodes))) + 'd}.vti'

            m = 0
            for lv in range(self.attributes['toplev']):

                sm = rm[::-1] * 2**lv
                bw = base.size / sm
                cw = bw / self.attributes['ncells']
                vtk.write('    <Block level="{}"'.format(lv) + \
                                ' spacing="{} {} {}">\n'.format(*cw))

                no = 0
                for item in base.getNodesFromLevel(lv):
                    lo = numpy.array(item.index) * bm
                    up = lo + bm - 1
                    ll = numpy.stack((lo,up)).T.flatten()
                    if item.hasData:
                        vfile = os.path.join(opath, fmt.format(dataset_name, lv, no))
                        WriteVTK(vfile, label, item.data, \
                            origin = (item.lower[0], item.lower[1], item.lower[2]), \
                            spacing = (cw[0], cw[1], cw[2]), \
                            compression=compression, compression_level=compression_level)
                        vtk.write('      <DataSet index="{}"'.format(no) + \
                                    ' amr_box = "{} {} {} {} {} {}"'.format(*ll) + \
                                        ' file = "{}"></DataSet>\n'.format(vfile))
                        no += 1
                    else:
                        vtk.write('      <DataSet index="{}"'.format(no) + \
                                    ' amr_box = "{} {} {} {} {} {}"'.format(*ll) + \
                                        '></DataSet>\n')
                    m += 1

                    if progress:
                        sys.stdout.write('\r')
                        sys.stdout.write("Storing AMR block {} of {}".format(m, base.nodes))
                        sys.stdout.flush()

                vtk.write('    </Block>\n')
            vtk.write('  </vtkOverlappingAMR>\n')
            vtk.write('</VTKFile>')

        if (progress):
            sys.stdout.write('\n')
            sys.stdout.flush()



"""
================================================================================

  This file is part of the AMUN source code, a program to perform
  Newtonian or relativistic magnetohydrodynamical simulations on uniform or
  adaptive mesh.

  Copyright (C) 2018-2022 Grzegorz Kowal <grzegorz@amuncode.org>

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

================================================================================

 module: AMUN

  This module implements an interface class to read attributes, coordinates,
  and datasets stored in AmunH5 format snapshots.

--------------------------------------------------------------------------------
"""
from .amun import Amun

class AmunH5(Amun):
    """AMUN H5 snapshot class"""

    def __init_snapshot__(self):
        """
            Sets the data format after verifying if the snapshot is stored
            in AmunH5 format.
        """
        import h5py

        if not self.path_is_file:
            raise Exception("AmunH5 requires a file not directory as the argument!")

        with h5py.File(self.path, 'r') as h5:
            if 'codes' in h5.attrs:
                if h5.attrs['code'].astype(str) == "AMUN":
                    self.dataformat = 'AmunH5'
                else:
                    raise Exception("'{}' contains attribute 'code', but its content is not 'AMUN'!".format(self.path))
            elif 'attributes' in h5 and 'coordinates' in h5 and 'variables' in h5:
                self.dataformat = 'AmunH5'
            else:
                raise Exception("{} misses one of these groups: 'attributes', 'coordinates' or 'variables'!".format(self.path))
            if 'version' in h5.attrs:
                if h5.attrs['version'].dtype == 'float32':
                    self.version = int(h5.attrs['version'])


    def __fill_attributes__(self):
        """
            Reads attributes from the snapshot file and adds them to
            the attributes' dictionary.
        """
        import h5py
        import numpy

        exclude_list = ['nseeds', 'seeds', 'dblocks', 'nproc', 'dims', 'dtn', 'last_id', 'mblocks']

        with h5py.File(self.path, 'r') as h5:
            for aname in h5['attributes'].attrs:
                if not aname in exclude_list:
                    attr = h5['attributes'].attrs[aname]
                    if attr.dtype in [ 'float64', 'float32', 'int64', 'int32' ]:
                        if isinstance(attr, numpy.ndarray):
                            self.attributes[aname] = numpy.squeeze(attr).tolist()
                        else:
                            self.attributes[aname] = attr
                    else:
                        self.attributes[aname] = numpy.squeeze(attr).astype(str)

        if not 'nchunks' in self.attributes and 'nprocs' in self.attributes:
            self.attributes['nchunks'] = self.attributes['nprocs']
            del self.attributes['nprocs']
        if 'rdims' in self.attributes:
            self.attributes['xblocks'] = self.attributes['rdims'][0]
            self.attributes['yblocks'] = self.attributes['rdims'][1]
            if self.attributes['ndims'] == 3:
                self.attributes['zblocks'] = self.attributes['rdims'][2]
            del self.attributes['rdims']


    def __fill_variables__(self):
        """
            Reads the names of datasets stored in the snapshot and adds them
            to the variables' dictionary.
        """
        import h5py

        with h5py.File(self.path, 'r') as h5:
            for variable in h5['variables']:
                v = variable.strip()
                self.variables[v] = v


    def __fill_chunks__(self):
        """
            Retrieves metadata about datablocks stored in the snapshot's chunks
            and adds them to the chunks' dictionary.
        """
        import h5py, numpy, os

        self.chunkname = 'p{:06d}'.format(self.attributes['isnap']) + '_{:05d}.h5'
        for n in range(self.attributes['nchunks']):
            self.chunks[n] = dict()
            self.chunks[n]['filename'] = self.chunkname.format(n)
            cname = os.path.join(self.dirname, self.chunks[n]['filename'])
            if os.path.exists(cname):
                with h5py.File(cname, 'r') as h5:
                    self.chunks[n]['dblocks'] = numpy.squeeze(h5['attributes'].attrs['dblocks'])
                    if self.chunks[n]['dblocks'] > 0:
                        self.chunks[n]['levels'] = numpy.array(h5['coordinates']['levels'])
                        self.chunks[n]['bounds'] = self.__swap__(numpy.array(h5['coordinates']['bounds']))
                        self.chunks[n]['coords'] = self.__swap__(numpy.array(h5['coordinates']['coords']))
                    else:
                        self.chunks[n]['levels'] = None
                        self.chunks[n]['coords'] = None
                        self.chunks[n]['bounds'] = None
            else:
                raise Exception("Snapshot's chunk '{}' not present!".format(cname))


    def __swap__(self, arr):
        """
            Function swaps the array for version before 1.0.
        """
        pass
        import numpy

        if self.version < 1:
            return numpy.transpose(arr, numpy.roll(numpy.arange(arr.ndim), 1))
        else:
            return arr


    def __read_binary_data__(self, dataset_name, chunk_number):
        """
            Gets the dataset array from a given snapshot's chunk.
        """
        import h5py, numpy, os

        cname = os.path.join(self.dirname, self.chunks[chunk_number]['filename'])
        with h5py.File(cname, 'r') as h5:
            return self.__swap__(numpy.squeeze(numpy.array(h5['variables'][dataset_name])))



# -*- coding: utf-8 -*-

"""

Package AmunPy is the Python interface to handle datasets produced by
the AMUN code (https://amuncode.org).

The package is released under GNU General Public License v3.
See file LICENSE for more details.

"""

from .amunxml import *
from .amunh5 import *
from .amunh5_deprecated import *
from .integrals import *
from .vtkio import *

__all__ = [ 'AmunXML', 'AmunH5', 'WriteVTK', \
        'amun_attribute', 'amun_coordinate', 'amun_dataset', 'amun_dataset_vtk', 'amun_integrals' ]

__author__ = "Grzegorz Kowal"
__copyright__ = "Copyright 2018-2022 Grzegorz Kowal <grzegorz@amuncode.org>"
__version__ = "0.9.6"
__maintainer__ = "Grzegorz Kowal"
__email__ = "grzegorz@amuncode.org"



"""
================================================================================

  This file is part of the AMUN source code, a program to perform
  Newtonian or relativistic magnetohydrodynamical simulations on uniform or
  adaptive mesh.

  Copyright (C) 2018-2022 Grzegorz Kowal <grzegorz@amuncode.org>

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

================================================================================

 module: AMUNXML

  This module implements an interface class to read attributes, coordinates,
  and datasets stored in AmunXML format snapshots.

--------------------------------------------------------------------------------
"""
from .amun import Amun

class AmunXML(Amun):
    """AMUN XML snapshot class"""

    def __init_snapshot__(self):
        """
            Sets the data format after verifying if the snapshot is stored
            in AmunXML format.
        """
        import os
        import xml.etree.ElementTree as ET

        if not self.path_is_file:
            self.filename = "metadata.xml"

        mfile = os.path.join(self.dirname, self.filename)
        if os.path.exists(mfile):
            tree = ET.parse(mfile)
            root = tree.getroot()
            if self.version < 0:
                if 'version' in root.attrib:
                        self.version = int(float(root.attrib['version']))
            if root.tag == 'AMUNFile':
                self.dataformat = 'AmunXML'
            else:
                raise Exception("{} does not seem to be an AmunXML snapshot!".format(mfile))
        else:
            raise Exception("{} does not exist!".format(mfile))


    def __fill_attributes__(self):
        """
            Reads attributes from the snapshot file and adds them to
            the attributes' dictionary.
        """
        import os
        import xml.etree.ElementTree as ET

        exclude_list = ['nproc']

        tree = ET.parse(os.path.join(self.dirname, self.filename))
        root = tree.getroot()
        for child in root:
            if not child.tag == 'BinaryFiles':
                for item in child:
                    if item.tag == 'Attribute':
                        if not item.attrib['name'] in exclude_list:
                            if item.attrib['type'] == 'double':
                                self.attributes[item.attrib['name']] = float(item.text)
                            elif item.attrib['type'] == 'integer':
                                self.attributes[item.attrib['name']] = int(item.text)
                            else:
                                self.attributes[item.attrib['name']] = item.text

        if not 'nchunks' in self.attributes and 'nprocs' in self.attributes:
            self.attributes['nchunks'] = self.attributes['nprocs']
            del self.attributes['nprocs']

        if not 'zmin' in self.attributes:
            self.attributes['zmin'] = 0
        if not 'zmax' in self.attributes:
            self.attributes['zmax'] = 1
        if not 'zblocks' in self.attributes:
            self.attributes['zblocks'] = 1


    def __fill_variables__(self):
        """
            Reads the names of datasets stored in the snapshot and adds them
            to the variables' dictionary.
        """
        if 'variables' in self.attributes:
            for v in self.attributes['variables'].split():
                self.variables[v] = v
        else:
            raise Exception("No attribute 'variables' in {}!".format(self.filename))


    def __fill_chunks__(self):
        """
            Retrieves metadata about datablocks stored in the snapshot's chunks
            and adds them to the chunks' dictionary.
        """
        import numpy, os
        import xml.etree.ElementTree as ET

        self.binaries = dict()

        fname = os.path.join(self.dirname, self.filename)
        tree  = ET.parse(fname)
        root  = tree.getroot()
        for child in root:
            if child.tag == 'BinaryFiles':
                for item in child:
                    self.binaries[item.attrib['name']]         = item.attrib
                    self.binaries[item.attrib['name']]['file'] = item.text

        self.chunkname = 'datablocks_{:06d}.xml'
        for n in range(self.attributes['nchunks']):
            self.chunks[n] = dict()
            self.chunks[n]['filename'] = self.chunkname.format(n)
            fname = os.path.join(self.dirname, self.chunks[n]['filename'])
            if os.path.exists(fname):
                tree = ET.parse(fname)
                root = tree.getroot()
                for item in root.findall('./BinaryFiles/Attribute'):
                    self.chunks[n][item.attrib['name']]         = item.attrib
                    self.chunks[n][item.attrib['name']]['file'] = item.text
                for item in root.findall('./DataBlocks/Attribute'):
                    if item.attrib['name'] == 'dblocks':
                        self.chunks[n]['dblocks'] = int(item.text)
            else:
                raise Exception("Snapshot's chunk '{}' does not exist!".format(fname))

        mset = self.__read_binary_meta('fields')

        n     = mset.size // self.attributes['nleafs']
        if self.version > 0:
            mset  = numpy.reshape(mset, [self.attributes['nleafs'],n])
        else:
            mset  = self.__swap__(numpy.reshape(mset, [n, self.attributes['nleafs']]))

        index = dict()
        level = dict()
        coord = dict()
        for n in range(self.attributes['nleafs']):
            index[mset[n,0]] = n
            level[mset[n,0]] = mset[n,  1]
            coord[mset[n,0]] = mset[n,2:5]

        bounds = self.__read_binary_meta('bounds', dtype='float64')
        if self.version > 0:
            bounds = numpy.reshape(bounds, [self.attributes['nleafs'], 2, 3])
        else:
            bounds = self.__swap__(numpy.reshape(bounds, [2, 3, self.attributes['nleafs']]))

        for n in range(self.attributes['nchunks']):
            nblk = self.chunks[n]['dblocks']
            if nblk > 0:
                ids = self.__read_binary_data('ids', n, dtype='int32')

                self.chunks[n]['levels'] = numpy.array([level[ids[p]] for p in range(nblk)])
                self.chunks[n]['coords'] = numpy.array([coord[ids[p]] for p in range(nblk)])

                ii  = [ index[ids[p]] for p in range(nblk) ]

                self.chunks[n]['bounds'] = numpy.array([bounds[ii[p],:,:] for p in range(nblk)])
            else:
                self.chunks[n]['levels'] = None
                self.chunks[n]['coords'] = None
                self.chunks[n]['bounds'] = None


    def __swap__(self, arr):
        """
            Function swaps the array for version before 1.0.
        """
        pass
        import numpy

        if self.version < 1:
            return numpy.transpose(arr, numpy.roll(numpy.arange(arr.ndim), 1))
        else:
            return arr


    def __read_binary_data__(self, dataset_name, chunk_number):
        """
            Gets the dataset array from a given snapshot's chunk.
        """
        import numpy

        dims = self.chunks[chunk_number]['dims']
        if self.version < 1:
            dims = numpy.roll(self.chunks[chunk_number]['dims'], -1)
        dset = numpy.reshape(self.__read_binary_data(dataset_name, chunk_number), dims)
        return self.__swap__(dset)


    def __check_digest(self, filename, hash_type, digest, data):
        '''
            Verifies if the provided digest matches the data.
        '''
        import xxhash

        failed = False
        if hash_type == 'xxh64':
            failed = digest.lower() != xxhash.xxh64(data).hexdigest()
        elif hash_type == 'xxh3':
            failed = digest.lower() != xxhash.xxh3_64(data).hexdigest()
        if failed:
            print("File '{}' seems to be corrupted! Proceeding anyway...".format(filename))


    def __read_binary_meta(self, dataset, dtype='int32'):
        '''
            Reads binary data of metadata.
        '''
        import numpy, os
        import zstandard as zstd
        import lz4.frame as lz4
        import lzma

        fname = os.path.join(self.path, self.binaries[dataset]['file'])
        if os.path.exists(fname):
            with open(fname, mode ='rb') as f:
                stream = f.read()

            if 'compression_format' in self.binaries[dataset]:
                if 'compressed_digest' in self.binaries[dataset]:
                    htype = self.binaries[dataset]['digest_type']
                    dhash = self.binaries[dataset]['compressed_digest']
                    self.__check_digest(fname, htype, dhash, stream)

                comp = self.binaries[dataset]['compression_format']
                if comp == 'zstd':
                    data = zstd.ZstdDecompressor().decompress(stream)
                elif comp == 'lz4':
                    data = lz4.decompress(stream)
                elif comp == 'lzma':
                    data = lzma.decompress(stream)
                else:
                    raise Exception("Binary file '{}' compressed in unsupported format {}!".format(fname, comp))

                if 'digest' in self.binaries[dataset]:
                    htype = self.binaries[dataset]['digest_type']
                    dhash = self.binaries[dataset]['digest']
                    self.__check_digest(fname, htype, dhash, data)

                return numpy.frombuffer(data, dtype=dtype)
            else:
                if 'digest' in self.binaries[dataset]:
                    htype = self.binaries[dataset]['digest_type']
                    dhash = self.binaries[dataset]['digest']
                    self.__check_digest(fname, htype, dhash, stream)

                return numpy.frombuffer(stream, dtype=dtype)
        else:
            raise Exception("Binary file '{}' does not exist!".format(fname))


    def __read_binary_data(self, dataset_name, chunk_number, dtype='float64'):
        '''
            Reads binary data of provided dataset name from a given chunk.
        '''
        import numpy, os
        import zstandard as zstd
        import lz4.frame as lz4
        import lzma

        fname = os.path.join(self.path, self.chunks[chunk_number][dataset_name]['file'])
        if os.path.exists(fname):
            with open(fname, mode ='rb') as f:
                stream = f.read()

            if 'compression_format' in self.chunks[chunk_number][dataset_name]:
                if 'compressed_digest' in self.chunks[chunk_number][dataset_name]:
                    htype = self.chunks[chunk_number][dataset_name]['digest_type']
                    dhash = self.chunks[chunk_number][dataset_name]['compressed_digest']
                    self.__check_digest(fname, htype, dhash, stream)

                comp = self.chunks[chunk_number][dataset_name]['compression_format']
                if comp == 'zstd':
                    dctx = zstd.ZstdDecompressor()
                    data = dctx.decompress(stream)
                elif comp == 'lz4':
                    data = lz4.decompress(stream)
                elif comp == 'lzma':
                    data = lzma.decompress(stream)
                else:
                    raise Exception("Binary file '{}' compressed in unsupported format {}!".format(fname, comp))

                if 'digest' in self.chunks[chunk_number][dataset_name]:
                    htype = self.chunks[chunk_number][dataset_name]['digest_type']
                    dhash = self.chunks[chunk_number][dataset_name]['digest']
                    self.__check_digest(fname, htype, dhash, data)

                return numpy.frombuffer(data, dtype=dtype)
            else:
                if 'digest' in self.chunks[chunk_number][dataset_name]:
                    htype = self.chunks[chunk_number][dataset_name]['digest_type']
                    dhash = self.chunks[chunk_number][dataset_name]['digest']
                    self.__check_digest(fname, htype, dhash, stream)

                return numpy.frombuffer(stream, dtype=dtype)
        else:
            raise Exception("Binary file '{}' does not exist!".format(fname))



"""
================================================================================

  This file is part of the AMUN source code, a program to perform
  Newtonian or relativistic magnetohydrodynamical simulations on uniform or
  adaptive mesh.

  Copyright (C) 2018-2022 Grzegorz Kowal <grzegorz@amuncode.org>

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

================================================================================

 module: AMUN

  Python module with subroutines to read AMUN code HDF5 files.

  The only requirements for this package are:

     - numpy

--------------------------------------------------------------------------------
"""
def amun_integrals(field, filename, pathlist):
    '''
        get_integral: iterate over pathlist and read and merge field values from filename files in the provided paths
    '''
    # Initiate the return values with empty array and file number.
    #
    vals = np.array([])
    num  = 1

    # Iterate over all paths provided in the list 'pathlist'.
    #
    for path in pathlist:

      # Iterate over all files in the current path.
      #
      while True:

        # Generate file name.
        #
        dfile = path + '/' + filename + '_' + str(num).zfill(2) + '.dat'

        # Check if the file exists.
        #
        if op.isfile(dfile):

          # Read values from the current integrals file.
          #
          lvals = read_integrals(dfile, field)

          # Append to the return array.
          #
          vals = np.append(vals, lvals)

          # Increase the number file.
          #
          num = num + 1

        else:

          # File does not exists, so go to the next path.
          #
          break

    # Return appended values.
    #
    return vals


def read_integrals(filename, column):
    '''
        read_integrals: reads a given column from an integral file.
    '''
    # Open the given file and check if it is text file.
    #
    f = open(filename, 'r')

    # Read fist line and store it in h, since it will be used to obtain the
    # column headers.
    #
    l = f.readline()
    h = l

    # Read first line which is not comment in order to determine the number of
    # columns and store the number of columns in nc.  Calculate the column width
    # and store it in wc.
    #
    while l.startswith('#'):
        l = f.readline()
    nc = len(l.rsplit())
    wc = int((len(l) - 9) / (nc - 1))

    # Split header line into a list.
    #
    lh = [h[1:9].strip()]
    for i in range(nc - 1):
      ib = i * wc + 10
      ie = ib + wc - 1
      lh.append(h[ib:ie].strip())


    ic = lh.index(column)

    # Read given column.
    #
    if (ic > -1):
      lc = [float(l.split()[ic])]
      for l in f:
        lc.append(float(l.split()[ic]))

    # Close the file.
    #
    f.close()

    # Return values.
    #
    return(np.array(lc))



"""
================================================================================

  This file is part of the AMUN source code, a program to perform
  Newtonian or relativistic magnetohydrodynamical simulations on uniform or
  adaptive mesh.

  Copyright (C) 2018-2022 Grzegorz Kowal <grzegorz@amuncode.org>

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

================================================================================

 module: VTKIO

  Module provides a function to store given dataset in the VTK image file.

--------------------------------------------------------------------------------
"""

def WriteVTK(vtkfile, vname, data, \
            origin=(0, 0, 0), spacing=(1, 1, 1), \
            compression=None, compression_level=19, encode=True, \
            lz4mode='default', lz4acceleration=1, block_size=32768, \
            points=False, verbose=False):

    import base64, numpy, struct, zlib, lz4.block, lzma

    if isinstance(data, (list, tuple)):

        if not all(isinstance(d, (numpy.ndarray)) for d in data):
            raise Exception('All input data components in WriteVTK must be arrays!')

        if not all(data[0].shape == d.shape for d in data):
            raise Exception('All input data components in WriteVTK must have the same dimensions!')

        dtype = 'vector'
        ncomp = len(data)
        ndims = data[0].ndim
        dims  = data[0].shape

    elif isinstance(data, (numpy.ndarray)):

        dtype = 'scalar'
        ncomp = 1
        ndims = data.ndim
        dims  = data.shape

    else:
        raise Exception('Unknown type of the input data in WriteVTK!')

    with open(vtkfile, 'wb') as vt:
        offset = 0
        sdims  = '"'
        if points:
            for i in range(ndims - 1, 0, -1):
                sdims +='%d %d ' % (0, dims[i] - 1)
            sdims += '%d %d" ' % (0, dims[0] - 1)
        else:
            for i in range(ndims - 1, 0, -1):
                sdims +='%d %d ' % (0, dims[i])
            sdims += '%d %d" ' % (0, dims[0])

        string  = '<?xml version="1.0"?>\n<VTKFile type="ImageData" version="1.0" byte_order="LittleEndian" header_type="UInt64"'
        if compression == 'zlib':
            string  += ' compressor="vtkZLibDataCompressor"'
            compression_level = min(max(compression_level, 0), 9)
        elif compression == 'lzma':
            string  += ' compressor="vtkLZMADataCompressor"'
            compression_level = min(max(compression_level, 0), 9)
        elif compression == 'lz4':
            string  += ' compressor="vtkLZ4DataCompressor"'
            compression_level = min(max(compression_level, 0), 12)
            lz4acceleration = max(lz4acceleration, 1)
        string += '>\n  <ImageData WholeExtent=%s' % sdims
        string += 'Origin="%e %e %e" ' % origin
        string += 'Spacing="%e %e %e">\n' % spacing
        string += '    <Piece Extent=%s>\n' % sdims
        if points:
            string += '      <PointData %ss="%s">\n' % (dtype, vname)
        else:
            string += '      <CellData %ss="%s">\n' % (dtype, vname)
        if ncomp == 1:
            dmin = data.min()
            dmax = data.max()
        else:
            dd   = numpy.zeros(data[0].shape)
            for dc in data:
                dd += dc**2
            dmin = numpy.sqrt(dd.min())
            dmax = numpy.sqrt(dd.max())
        string += '        <DataArray '
        if ncomp > 1:
            string += 'NumberOfComponents="{:d}" '.format(ncomp)
        string += 'type="Float32" Name="{}" RangeMin="{:e}" RangeMax="{:e}" format="appended" offset="{}">\n'.format(vname, dmin, dmax, offset)
        string += "        </DataArray>\n"
        if points:
            string += '      </PointData>\n'
        else:
            string += '      </CellData>\n'
        string += '    </Piece>\n'
        string += '  </ImageData>\n'
        vt.write(string.encode())

        if encode:
            string  = '  <AppendedData encoding="base64">\n'
        else:
            string  = '  <AppendedData encoding="raw">\n'
        string += '   _'
        vt.write(string.encode())

        if dtype == 'vector':
            qt = numpy.zeros([ dims[0], dims[1], dims[2], ncomp ], dtype=numpy.float32)
            for n, d in enumerate(data[:]):
                qt[:,:,:,n] = numpy.float32(d)
        else:
            qt = numpy.float32(data)

        barr = qt.tobytes()

        if compression != None:
            if len(barr) < block_size:
                if compression == 'zlib':
                    carr = zlib.compress(barr, compression_level)
                elif compression == 'lz4':
                    carr = lz4.block.compress(barr, mode=lz4mode, acceleration=lz4acceleration, compression=compression_level, store_size=False)
                elif compression == 'lzma':
                    carr = lzma.compress(barr)

                head = struct.pack("QQQQ", 1, len(barr), 0, len(carr))

            else:
                nblocks = len(barr) // block_size
                rsize   = len(barr) % block_size
                if verbose:
                    print("Number of blocks:\t {}\nBlock size:\t\t {}\nRemaining size:\t\t {}".format(nblocks, block_size, rsize))

                head = struct.pack("QQQ", nblocks, block_size, rsize)

                carr = bytearray(b'')
                ib = 0
                if compression == 'zlib':
                    cctx = zlib
                    for i in range(nblocks):
                        ie = min(len(barr), ib + block_size)
                        cblk = cctx.compress(barr[ib:ie], compression_level)
                        head += struct.pack("Q", len(cblk))
                        carr += cblk
                        ib = ie

                elif compression == 'lzma':
                    cctx = lzma
                    for i in range(nblocks):
                        ie = min(len(barr), ib + block_size)
                        cblk = cctx.compress(barr[ib:ie], preset=compression_level)
                        head += struct.pack("Q", len(cblk))
                        carr += cblk
                        ib = ie

                elif compression == 'lz4':
                    cctx = lz4.block
                    for i in range(nblocks):
                        ie = min(len(barr), ib + block_size)
                        cblk = cctx.compress(barr[ib:ie], mode=lz4mode, acceleration=lz4acceleration, compression=compression_level, store_size=False)
                        head += struct.pack("Q", len(cblk))
                        carr += cblk
                        ib = ie

            if encode:
                vt.write(base64.b64encode(head)+base64.b64encode(carr))
            else:
                vt.write(head+carr)

        else:
            head = struct.pack("Q", len(barr))
            if encode:
                vt.write(base64.b64encode(head+barr))
            else:
                vt.write(head+barr)

        string  = '\n  </AppendedData>\n'
        string += '</VTKFile>\n'
        vt.write(string.encode())



"""
================================================================================

  This file is part of the AMUN source code, a program to perform
  Newtonian or relativistic magnetohydrodynamical simulations on uniform or
  adaptive mesh.

  Copyright (C) 2018-2022 Grzegorz Kowal <grzegorz@amuncode.org>

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

================================================================================

 module: AMUN

  Python module with subroutines to read AMUN code HDF5 files.

  The only requirements for this package are:

     - h5py
     - numpy

--------------------------------------------------------------------------------
"""
#===============================================================================
'''
  DEPRECATED FUNCTIONS
'''

def amun_compatible(fname):
  '''
      Subroutine checks if the HDF5 file is AMUN compatible.

      Arguments:

        fname - the HDF5 file name;

      Return values:

        True or False;

      Examples:

        comp = amun_compatible('p000010_00000.h5')

  '''
  from warnings import warn
  import h5py as h5

  warn('This function is deprecated', DeprecationWarning, stacklevel=2)

  with h5.File(fname, 'r') as f:
    if 'codes' in f.attrs:
      if f.attrs['code'].astype(str) == "AMUN":
        return True
      else:
        print("'%s' contains attribute 'code'," % fname, \
              " but it is not 'AMUN'!")
        return False
    elif 'attributes' in f and 'coordinates' in f and \
         'variables'  in f:
      return True
    else:
      print("'%s' misses one of these groups:" % fname, \
            "'attributes', 'coordinates' or 'variables'!")
      return False


def amun_attribute(fname, aname):
  '''
      Subroutine to read global attributes from AMUN HDF5 snapshots.

      Arguments:

        fname - the HDF5 file name;
        aname - the attribute name;

      Return values:

        ret   - the value of the attribute or None;

      Examples:

        time = amun_attribute('p000010_00000.h5', 'time')

  '''
  from warnings import warn
  import h5py as h5
  import numpy as np

  warn('This function is deprecated', DeprecationWarning, stacklevel=2)

  if not amun_compatible(fname):
    return None

  with h5.File(fname, 'r') as f:
    if aname in f['attributes'].attrs:
      attr = f['attributes'].attrs[aname]
      if attr.dtype.type is np.string_:
        ret = np.squeeze(attr).astype(str)
      else:
        ret = np.squeeze(attr)
      return ret
    else:
      print("Attribute '%s' cannot be found in '%s'!" % (aname, fname))
      return None


def amun_coordinate(fname, iname):
  '''
      Subroutine to read coordinate items from AMUN HDF5 snapshots.

      Arguments:

        fname - the HDF5 file name;
        iname - the item name;

      Return values:

        ret   - the value of the item or None;

      Examples:

        bounds = amun_coordinate('p000010_00000.h5', 'bounds')

  '''
  from warnings import warn
  import h5py as h5
  import numpy as np

  warn('This function is deprecated', DeprecationWarning, stacklevel=2)

  if not amun_compatible(fname):
    return None

  with h5.File(fname, 'r') as f:
    if iname in f['coordinates']:
      return np.array(f['coordinates'][iname])
    else:
      print("Coordinate item '%s' not found in group 'coordinate' of '%s'!" % (iname, fname))
      return None


def amun_dataset(fname, vname, shrink=1, interpolation='rebin', order=3, progress=False):
  '''
      Subroutine to read datasets from AMUN HDF5 snapshots.

      Arguments:

        fname    - the HDF5 file name;
        vname    - the variable name;
        shrink   - the shrink factor (must be the power of 2 and not larger
                   than the block size);
        progress - the progress bar switch;

      Return values:

        ret   - the array of values for the variable;

      Examples:

        dn = amun_dataset('p000010_00000.h5', 'dens')

  '''
  from .interpolation import interpolate
  from warnings import warn
  import h5py as h5
  import numpy as np
  import os, sys

  warn('This function is deprecated', DeprecationWarning, stacklevel=2)

  if not amun_compatible(fname):
    return None

  dname = os.path.dirname(fname)

  if progress:
    sys.stdout.write("Data file path:\n  '%s'\n" % (dname))

  # get attributes necessary to reconstruct the domain
  #
  eqsys = amun_attribute(fname, 'eqsys')
  eos   = amun_attribute(fname, 'eos')
  nr    = amun_attribute(fname, 'isnap')
  nc    = amun_attribute(fname, 'nprocs')
  nl    = amun_attribute(fname, 'nleafs')
  if eos == 'adi':
    gm  = amun_attribute(fname, 'adiabatic_index')

  # get block dimensions and the maximum level
  #
  ndims = amun_attribute(fname, 'ndims')
  nn    = amun_attribute(fname, 'ncells')
  bm    = np.array([nn, nn, nn])
  if ndims == 2:
    bm[2] = 1
  ng    = amun_attribute(fname, 'nghosts')
  ml    = amun_attribute(fname, 'maxlev')

  # get the base block dimensions
  #
  rm    = amun_attribute(fname, 'bdims')
  if rm is None:
    rm  = amun_attribute(fname, 'domain_base_dims')
  if rm is None:
    rm  = amun_attribute(fname, 'rdims')
  if rm is None:
    return None

  # build the list of supported variables
  #
  variables = []
  with h5.File(fname, 'r') as f:
    for var in f['variables'].keys():
      variables.append(var)

  # add derived variables if possible
  #
  variables.append('level')
  if 'velx' in variables and 'vely' in variables and 'velz' in variables:
    variables.append('velo')
    variables.append('divv')
    variables.append('vort')
  if 'magx' in variables and 'magy' in variables and 'magz' in variables:
    variables.append('magn')
    variables.append('divb')
    variables.append('curr')
  if (eqsys == 'hd' or eqsys == 'mhd') and eos == 'adi' \
                    and 'pres' in variables:
    variables.append('eint')
  if 'dens' in variables and 'pres' in variables:
    variables.append('temp')
  if (eqsys == 'hd' or eqsys == 'mhd') \
                    and 'dens' in variables \
                    and 'velx' in variables \
                    and 'vely' in variables \
                    and 'velz' in variables:
    variables.append('ekin')
  if (eqsys == 'mhd' or eqsys == 'srmhd') \
                     and 'magx' in variables \
                     and 'magy' in variables \
                     and 'magz' in variables:
    variables.append('emag')
  if eqsys == 'hd' and 'ekin' in variables and 'eint' in variables:
    variables.append('etot')
  if eqsys == 'mhd' and 'eint' in variables \
                    and 'ekin' in variables \
                    and 'emag' in variables:
    variables.append('etot')
  if (eqsys == 'srhd' or eqsys == 'srmhd') and 'velo' in variables:
    variables.append('lore')

  # check if the requested variable is in the variable list
  #
  if not vname in variables:
    print('The requested variable cannot be extracted from the file datasets!')
    return None

  # check if the shrink parameter is correct (block dimensions should be
  # divisible by the shrink factor)
  #
  shrink = max(1, int(shrink))
  if shrink > 1:
    if (nn % shrink) != 0:
      print('The block dimension should be divisible by the shrink factor!')
      return None
    sh = shrink
    while(sh > 2 and sh % 2 == 0):
      sh = int(sh / 2)
    if (sh % 2) != 0:
      print('The shrink factor should be a power of 2!')
      return None

  # determine the actual maximum level from the blocks
  #
  levs = []
  for n in range(nc):
    fname = 'p%06d_%05d.h5' % (nr, n)
    lname = os.path.join(dname, fname)
    dblocks = amun_attribute(lname, 'dblocks')
    if dblocks > 0:
      levs = np.append(levs, [amun_coordinate(lname, 'levels')])
  ml = int(levs.max())

  # prepare dimensions of the output array and allocate it
  #
  dm = np.array(rm[0:ndims] * bm[0:ndims] * 2**(ml - 1) / shrink, \
                                                        dtype=np.int32)
  ret = np.zeros(dm[::-1])

  # iterate over all subdomain files
  #
  nb = 0
  for n in range(nc):
    fname = 'p%06d_%05d.h5' % (nr, n)
    lname = os.path.join(dname, fname)
    dblocks = amun_attribute(lname, 'dblocks')
    if dblocks > 0:
      levels = amun_coordinate(lname, 'levels')
      coords = amun_coordinate(lname, 'coords')
      dx     = amun_coordinate(lname, 'dx')
      dy     = amun_coordinate(lname, 'dy')
      dz     = amun_coordinate(lname, 'dz')
      with h5.File(lname, 'r') as f:
        g       = f['variables']
        if vname == 'level':
          dataset = np.zeros(g[variables[0]].shape)
          for l in range(dblocks):
            dataset[:,:,:,l] = levels[l]
        elif vname == 'velo':
          dataset = np.sqrt(g['velx'][:,:,:,:]**2 \
                          + g['vely'][:,:,:,:]**2 \
                          + g['velz'][:,:,:,:]**2)
        elif vname == 'magn':
          dataset = np.sqrt(g['magx'][:,:,:,:]**2 \
                          + g['magy'][:,:,:,:]**2 \
                          + g['magz'][:,:,:,:]**2)
        elif vname == 'eint':
          dataset = 1.0 / (gm - 1.0) * g['pres'][:,:,:,:]
        elif vname == 'ekin':
          dataset = 0.5 * g['dens'][:,:,:,:] * (g['velx'][:,:,:,:]**2 \
                                              + g['vely'][:,:,:,:]**2 \
                                              + g['velz'][:,:,:,:]**2)
        elif vname == 'emag':
          dataset = 0.5 * (g['magx'][:,:,:,:]**2 \
                         + g['magy'][:,:,:,:]**2 \
                         + g['magz'][:,:,:,:]**2)
        elif vname == 'etot':
          dataset = 1.0 / (gm - 1.0) * g['pres'][:,:,:,:] \
                  + 0.5 * g['dens'][:,:,:,:] * (g['velx'][:,:,:,:]**2 \
                                              + g['vely'][:,:,:,:]**2 \
                                              + g['velz'][:,:,:,:]**2)
          if eqsys == 'mhd':
            dataset += 0.5 * (g['magx'][:,:,:,:]**2 \
                            + g['magy'][:,:,:,:]**2 \
                            + g['magz'][:,:,:,:]**2)
        elif vname == 'temp':
          dataset = g['pres'][:,:,:,:] / g['dens'][:,:,:,:]
        elif vname == 'lore':
          dataset = 1.0 / np.sqrt(1.0 - (g['velx'][:,:,:,:]**2 \
                                       + g['vely'][:,:,:,:]**2 \
                                       + g['velz'][:,:,:,:]**2))
        elif vname == 'divv':
          dataset = np.zeros(g['velx'].shape)
          fields  = [ 'velx', 'vely', 'velz' ]
          h       = (dx, dy, dz)
          for i in range(ndims):
            v = fields[i]
            dataset += 0.5 * (np.roll(g[v][:,:,:,:], -1, axis=2)  \
                            - np.roll(g[v][:,:,:,:],  1, axis=2)) \
                                                      / h[i][levels[:] - 1]
        elif vname == 'divb':
          dataset = np.zeros(g['magx'].shape)
          fields  = [ 'magx', 'magy', 'magz' ]
          h       = (dx, dy, dz)
          for i in range(ndims):
            v = fields[i]
            dataset += 0.5 * (np.roll(g[v][:,:,:,:], -1, axis=2)  \
                            - np.roll(g[v][:,:,:,:],  1, axis=2)) \
                                                      / h[i][levels[:] - 1]
        elif vname == 'vort':
          if ndims == 3:
            wx = 0.5 * (np.roll(g['velz'][:,:,:,:], -1, axis=1)  \
                      - np.roll(g['velz'][:,:,:,:],  1, axis=1)) \
                                                  / dy[levels[:]-1] \
               - 0.5 * (np.roll(g['vely'][:,:,:,:], -1, axis=0)  \
                      - np.roll(g['vely'][:,:,:,:],  1, axis=0)) \
                                                  / dz[levels[:]-1]
            wy = 0.5 * (np.roll(g['velx'][:,:,:,:], -1, axis=0)  \
                      - np.roll(g['velx'][:,:,:,:],  1, axis=0)) \
                                                  / dz[levels[:]-1] \
               - 0.5 * (np.roll(g['velz'][:,:,:,:], -1, axis=2)  \
                      - np.roll(g['velz'][:,:,:,:],  1, axis=2)) \
                                                  / dx[levels[:]-1]
            wz = 0.5 * (np.roll(g['vely'][:,:,:,:], -1, axis=2)  \
                      - np.roll(g['vely'][:,:,:,:],  1, axis=2)) \
                                                  / dx[levels[:]-1] \
               - 0.5 * (np.roll(g['velx'][:,:,:,:], -1, axis=1)  \
                      - np.roll(g['velx'][:,:,:,:],  1, axis=1)) \
                                                  / dy[levels[:]-1]
          else:
            wx =   0.5 * (np.roll(g['velz'][:,:,:,:], -1, axis=1)  \
                        - np.roll(g['velz'][:,:,:,:],  1, axis=1)) \
                                                   / dy[levels[:]-1]
            wy = - 0.5 * (np.roll(g['velz'][:,:,:,:], -1, axis=2)  \
                        - np.roll(g['velz'][:,:,:,:],  1, axis=2)) \
                                                   / dx[levels[:]-1]
            wz =   0.5 * (np.roll(g['vely'][:,:,:,:], -1, axis=2)  \
                        - np.roll(g['vely'][:,:,:,:],  1, axis=2)) \
                                                   / dx[levels[:]-1] \
                 - 0.5 * (np.roll(g['velx'][:,:,:,:], -1, axis=1)  \
                        - np.roll(g['velx'][:,:,:,:],  1, axis=1)) \
                                                   / dy[levels[:]-1]
          dataset = np.sqrt(wx * wx + wy * wy + wz * wz)
        elif vname == 'curr':
          if ndims == 3:
            wx = 0.5 * (np.roll(g['magz'][:,:,:,:], -1, axis=1)  \
                      - np.roll(g['magz'][:,:,:,:],  1, axis=1)) \
                                                  / dy[levels[:]-1] \
               - 0.5 * (np.roll(g['magy'][:,:,:,:], -1, axis=0)  \
                      - np.roll(g['magy'][:,:,:,:],  1, axis=0)) \
                                                  / dz[levels[:]-1]
            wy = 0.5 * (np.roll(g['magx'][:,:,:,:], -1, axis=0)  \
                      - np.roll(g['magx'][:,:,:,:],  1, axis=0)) \
                                                  / dz[levels[:]-1] \
               - 0.5 * (np.roll(g['magz'][:,:,:,:], -1, axis=2)  \
                      - np.roll(g['magz'][:,:,:,:],  1, axis=2)) \
                                                  / dx[levels[:]-1]
            wz = 0.5 * (np.roll(g['magy'][:,:,:,:], -1, axis=2)  \
                      - np.roll(g['magy'][:,:,:,:],  1, axis=2)) \
                                                  / dx[levels[:]-1] \
               - 0.5 * (np.roll(g['magx'][:,:,:,:], -1, axis=1)  \
                      - np.roll(g['magx'][:,:,:,:],  1, axis=1)) \
                                                  / dy[levels[:]-1]
          else:
            wx =   0.5 * (np.roll(g['magz'][:,:,:,:], -1, axis=1)  \
                        - np.roll(g['magz'][:,:,:,:],  1, axis=1)) \
                                                   / dy[levels[:]-1]
            wy = - 0.5 * (np.roll(g['magz'][:,:,:,:], -1, axis=2)  \
                        - np.roll(g['magz'][:,:,:,:],  1, axis=2)) \
                                                   / dx[levels[:]-1]
            wz =   0.5 * (np.roll(g['magy'][:,:,:,:], -1, axis=2)  \
                        - np.roll(g['magy'][:,:,:,:],  1, axis=2)) \
                                                   / dx[levels[:]-1] \
                 - 0.5 * (np.roll(g['magx'][:,:,:,:], -1, axis=1)  \
                        - np.roll(g['magx'][:,:,:,:],  1, axis=1)) \
                                                   / dy[levels[:]-1]
          dataset = np.sqrt(wx * wx + wy * wy + wz * wz)
        else:
          dataset = g[vname][:,:,:,:]

        # rescale all blocks to the effective resolution
        #
        for l in range(dblocks):
          nn   = 2**(ml - levels[l])
          if nn <= shrink:
            method = 'rebin'
          else:
            method = interpolation
          cm   = np.array(bm[0:ndims] * nn / shrink, dtype=np.int32)
          ibeg = coords[0:ndims,l] * cm[0:ndims]
          iend = ibeg + cm[0:ndims]
          if ndims == 3:
            ib, jb, kb = ibeg[0], ibeg[1], ibeg[2]
            ie, je, ke = iend[0], iend[1], iend[2]
            ret[kb:ke,jb:je,ib:ie] = interpolate(dataset[:,:,:,l], cm, ng, method=method, order=order)
          else:
            ib, jb = ibeg[0], ibeg[1]
            ie, je = iend[0], iend[1]
            ret[jb:je,ib:ie]       = interpolate(dataset[0,:,:,l], cm, ng, method=method, order=order)

          nb += 1

          # print progress bar if desired
          #
          if progress:
            sys.stdout.write('\r')
            sys.stdout.write("Reading '%s' from '%s': block %d from %d" \
                          % (vname, fname, nb, nl))
            sys.stdout.flush()

  if (progress):
    sys.stdout.write('\n')
    sys.stdout.flush()

  return ret


def amun_dataset_vtk(fname, vname, label=None, compression=None, compression_level=19, progress=False):
    '''
        Subroutine to convert a dataset specified by argument 'vname' from
        the AMUN HDF5 snapshot to OverlappedAMR VTK file.

        Arguments:

          fname       - the HDF5 file name;
          vname       - the variable name;
          label       - the variable label (long name);
          compression - the compression type: 'lz4', 'zlib', 'lzma'
          progress    - the progress bar switch;

        Examples:

          dn = amun_dataset_vtk('p000010_00000.h5', 'dens')

    '''
    from .octree import OcBase, OcNode
    from .vtkio import WriteVTK
    from warnings import warn
    import numpy as np
    import os, sys

    warn('This function is deprecated', DeprecationWarning, stacklevel=2)

    if not amun_compatible(fname):
        return None

    if amun_attribute(fname, 'ndims') < 3:
        print('Subroutine amun_dataset_vtk() supports only 3D domains.')
        return None

    if label == None:
        label = vname

    dname = os.path.dirname(fname)

    if progress:
        sys.stdout.write("Data file path:\n  '%s'\n" % (dname))

    # get attributes necessary to reconstruct the domain
    #
    eqsys = amun_attribute(fname, 'eqsys')
    eos   = amun_attribute(fname, 'eos')
    nr    = amun_attribute(fname, 'isnap')
    nc    = amun_attribute(fname, 'nprocs')
    nl    = amun_attribute(fname, 'nleafs')
    if eos == 'adi':
        gm  = amun_attribute(fname, 'adiabatic_index')

    # get block dimensions and the maximum level
    #
    ndims = amun_attribute(fname, 'ndims')
    nn    = amun_attribute(fname, 'ncells')
    bm    = np.array([nn, nn, nn])
    ng    = amun_attribute(fname, 'nghosts')
    ml    = amun_attribute(fname, 'maxlev')

    # get the base block dimensions
    #
    rm    = amun_attribute(fname, 'bdims')
    if rm is None:
        rm  = amun_attribute(fname, 'domain_base_dims')
    if rm is None:
        rm  = amun_attribute(fname, 'rdims')
    if rm is None:
        return None

    # get domain bounds
    #
    xmin = amun_attribute(fname, 'xmin')
    ymin = amun_attribute(fname, 'ymin')
    zmin = amun_attribute(fname, 'zmin')
    xlen = amun_attribute(fname, 'xmax') - xmin
    ylen = amun_attribute(fname, 'ymax') - ymin
    zlen = amun_attribute(fname, 'zmax') - zmin

    # build the list of supported variables
    #
    variables = []
    with h5.File(fname, 'r') as f:
      for var in f['variables'].keys():
        variables.append(var)

    # add derived variables if possible
    #
    variables.append('level')
    if 'velx' in variables and 'vely' in variables and 'velz' in variables:
        variables.append('velo')
        variables.append('divv')
        variables.append('vort')
    if 'magx' in variables and 'magy' in variables and 'magz' in variables:
        variables.append('magn')
        variables.append('divb')
        variables.append('curr')
    if (eqsys == 'hd' or eqsys == 'mhd') and eos == 'adi' \
                      and 'pres' in variables:
        variables.append('eint')
    if 'dens' in variables and 'pres' in variables:
        variables.append('temp')
    if (eqsys == 'hd' or eqsys == 'mhd') \
                      and 'dens' in variables \
                      and 'velx' in variables \
                      and 'vely' in variables \
                      and 'velz' in variables:
        variables.append('ekin')
    if (eqsys == 'mhd' or eqsys == 'srmhd') \
                       and 'magx' in variables \
                       and 'magy' in variables \
                       and 'magz' in variables:
        variables.append('emag')
    if eqsys == 'hd' and 'ekin' in variables and 'eint' in variables:
        variables.append('etot')
    if eqsys == 'mhd' and 'eint' in variables \
                      and 'ekin' in variables \
                      and 'emag' in variables:
        variables.append('etot')
    if (eqsys == 'srhd' or eqsys == 'srmhd') and 'velo' in variables:
        variables.append('lore')

    # check if the requested variable is in the variable list
    #
    if not vname in variables:
        print('The requested variable cannot be extracted from the file datasets!')
        return None

    # determine the actual maximum level from the blocks
    #
    levs = []
    for n in range(nc):
        fname = 'p%06d_%05d.h5' % (nr, n)
        lname = os.path.join(dname, fname)
        dblocks = amun_attribute(lname, 'dblocks')
        if dblocks > 0:
            levs = np.append(levs, [amun_coordinate(lname, 'levels')])
    ml = int(levs.max())

    # create octree base
    base = OcBase([xmin, ymin, zmin], [xlen, ylen, zlen], rm)

    # iterate over all subdomain files
    #
    nb = 0
    for n in range(nc):
        fname = 'p%06d_%05d.h5' % (nr, n)
        lname = os.path.join(dname, fname)
        dblocks = amun_attribute(lname, 'dblocks')
        if dblocks > 0:
            levels = amun_coordinate(lname, 'levels')
            coords = amun_coordinate(lname, 'coords')
            bounds = amun_coordinate(lname, 'bounds')
            dx     = amun_coordinate(lname, 'dx')
            dy     = amun_coordinate(lname, 'dy')
            dz     = amun_coordinate(lname, 'dz')
            with h5.File(lname, 'r') as f:
                g = f['variables']
                if vname == 'level':
                    dataset = np.zeros(g[variables[0]].shape)
                    for l in range(dblocks):
                        dataset[:,:,:,l] = levels[l]
                elif vname == 'velo':
                    dataset = np.sqrt(g['velx'][:,:,:,:]**2 \
                                    + g['vely'][:,:,:,:]**2 \
                                    + g['velz'][:,:,:,:]**2)
                elif vname == 'magn':
                    dataset = np.sqrt(g['magx'][:,:,:,:]**2 \
                                    + g['magy'][:,:,:,:]**2 \
                                    + g['magz'][:,:,:,:]**2)
                elif vname == 'eint':
                    dataset = 1.0 / (gm - 1.0) * g['pres'][:,:,:,:]
                elif vname == 'ekin':
                    dataset = 0.5 * g['dens'][:,:,:,:] * (g['velx'][:,:,:,:]**2 \
                                                        + g['vely'][:,:,:,:]**2 \
                                                        + g['velz'][:,:,:,:]**2)
                elif vname == 'emag':
                    dataset = 0.5 * (g['magx'][:,:,:,:]**2 \
                                   + g['magy'][:,:,:,:]**2 \
                                   + g['magz'][:,:,:,:]**2)
                elif vname == 'etot':
                    dataset = 1.0 / (gm - 1.0) * g['pres'][:,:,:,:] \
                            + 0.5 * g['dens'][:,:,:,:] * (g['velx'][:,:,:,:]**2 \
                                                        + g['vely'][:,:,:,:]**2 \
                                                        + g['velz'][:,:,:,:]**2)
                    if eqsys == 'mhd':
                      dataset += 0.5 * (g['magx'][:,:,:,:]**2 \
                                      + g['magy'][:,:,:,:]**2 \
                                      + g['magz'][:,:,:,:]**2)
                elif vname == 'temp':
                    dataset = g['pres'][:,:,:,:] / g['dens'][:,:,:,:]
                elif vname == 'lore':
                    dataset = 1.0 / np.sqrt(1.0 - (g['velx'][:,:,:,:]**2 \
                                                 + g['vely'][:,:,:,:]**2 \
                                                 + g['velz'][:,:,:,:]**2))
                elif vname == 'divv':
                    dataset = np.zeros(g['velx'].shape)
                    fields  = [ 'velx', 'vely', 'velz' ]
                    h       = (dx, dy, dz)
                    for i in range(ndims):
                      v = fields[i]
                      dataset += 0.5 * (np.roll(g[v][:,:,:,:], -1, axis=2)  \
                                      - np.roll(g[v][:,:,:,:],  1, axis=2)) \
                                                                / h[i][levels[:] - 1]
                elif vname == 'divb':
                    dataset = np.zeros(g['magx'].shape)
                    fields  = [ 'magx', 'magy', 'magz' ]
                    h       = (dx, dy, dz)
                    for i in range(ndims):
                      v = fields[i]
                      dataset += 0.5 * (np.roll(g[v][:,:,:,:], -1, axis=2)  \
                                      - np.roll(g[v][:,:,:,:],  1, axis=2)) \
                                                                / h[i][levels[:] - 1]
                elif vname == 'vort':
                    if ndims == 3:
                        wx = 0.5 * (np.roll(g['velz'][:,:,:,:], -1, axis=1)  \
                                  - np.roll(g['velz'][:,:,:,:],  1, axis=1)) \
                                                              / dy[levels[:]-1] \
                           - 0.5 * (np.roll(g['vely'][:,:,:,:], -1, axis=0)  \
                                  - np.roll(g['vely'][:,:,:,:],  1, axis=0)) \
                                                              / dz[levels[:]-1]
                        wy = 0.5 * (np.roll(g['velx'][:,:,:,:], -1, axis=0)  \
                                  - np.roll(g['velx'][:,:,:,:],  1, axis=0)) \
                                                              / dz[levels[:]-1] \
                           - 0.5 * (np.roll(g['velz'][:,:,:,:], -1, axis=2)  \
                                  - np.roll(g['velz'][:,:,:,:],  1, axis=2)) \
                                                              / dx[levels[:]-1]
                        wz = 0.5 * (np.roll(g['vely'][:,:,:,:], -1, axis=2)  \
                                  - np.roll(g['vely'][:,:,:,:],  1, axis=2)) \
                                                              / dx[levels[:]-1] \
                           - 0.5 * (np.roll(g['velx'][:,:,:,:], -1, axis=1)  \
                                  - np.roll(g['velx'][:,:,:,:],  1, axis=1)) \
                                                              / dy[levels[:]-1]
                    else:
                        wx =   0.5 * (np.roll(g['velz'][:,:,:,:], -1, axis=1)  \
                                    - np.roll(g['velz'][:,:,:,:],  1, axis=1)) \
                                                               / dy[levels[:]-1]
                        wy = - 0.5 * (np.roll(g['velz'][:,:,:,:], -1, axis=2)  \
                                    - np.roll(g['velz'][:,:,:,:],  1, axis=2)) \
                                                               / dx[levels[:]-1]
                        wz =   0.5 * (np.roll(g['vely'][:,:,:,:], -1, axis=2)  \
                                    - np.roll(g['vely'][:,:,:,:],  1, axis=2)) \
                                                               / dx[levels[:]-1] \
                             - 0.5 * (np.roll(g['velx'][:,:,:,:], -1, axis=1)  \
                                    - np.roll(g['velx'][:,:,:,:],  1, axis=1)) \
                                                               / dy[levels[:]-1]
                    dataset = np.sqrt(wx * wx + wy * wy + wz * wz)
                elif vname == 'curr':
                    if ndims == 3:
                        wx = 0.5 * (np.roll(g['magz'][:,:,:,:], -1, axis=1)  \
                                  - np.roll(g['magz'][:,:,:,:],  1, axis=1)) \
                                                              / dy[levels[:]-1] \
                           - 0.5 * (np.roll(g['magy'][:,:,:,:], -1, axis=0)  \
                                  - np.roll(g['magy'][:,:,:,:],  1, axis=0)) \
                                                              / dz[levels[:]-1]
                        wy = 0.5 * (np.roll(g['magx'][:,:,:,:], -1, axis=0)  \
                                  - np.roll(g['magx'][:,:,:,:],  1, axis=0)) \
                                                              / dz[levels[:]-1] \
                           - 0.5 * (np.roll(g['magz'][:,:,:,:], -1, axis=2)  \
                                  - np.roll(g['magz'][:,:,:,:],  1, axis=2)) \
                                                              / dx[levels[:]-1]
                        wz = 0.5 * (np.roll(g['magy'][:,:,:,:], -1, axis=2)  \
                                  - np.roll(g['magy'][:,:,:,:],  1, axis=2)) \
                                                              / dx[levels[:]-1] \
                           - 0.5 * (np.roll(g['magx'][:,:,:,:], -1, axis=1)  \
                                  - np.roll(g['magx'][:,:,:,:],  1, axis=1)) \
                                                              / dy[levels[:]-1]
                    else:
                        wx =   0.5 * (np.roll(g['magz'][:,:,:,:], -1, axis=1)  \
                                    - np.roll(g['magz'][:,:,:,:],  1, axis=1)) \
                                                               / dy[levels[:]-1]
                        wy = - 0.5 * (np.roll(g['magz'][:,:,:,:], -1, axis=2)  \
                                    - np.roll(g['magz'][:,:,:,:],  1, axis=2)) \
                                                               / dx[levels[:]-1]
                        wz =   0.5 * (np.roll(g['magy'][:,:,:,:], -1, axis=2)  \
                                    - np.roll(g['magy'][:,:,:,:],  1, axis=2)) \
                                                               / dx[levels[:]-1] \
                             - 0.5 * (np.roll(g['magx'][:,:,:,:], -1, axis=1)  \
                                    - np.roll(g['magx'][:,:,:,:],  1, axis=1)) \
                                                               / dy[levels[:]-1]
                    dataset = np.sqrt(wx * wx + wy * wy + wz * wz)
                else:
                    dataset = g[vname][:,:,:,:]

                # rescale all blocks to the effective resolution
                #
                for l in range(dblocks):

                    lv = levels[l] - 1

                    center = (bounds[0,:,l] + bounds[1,:,l]) / 2
                    base.createNodeBranch(center, lv)
                    base.setNodeData(center, lv, dataset[ng:-ng,ng:-ng,ng:-ng,l])

                    nb += 1

                    # print progress bar if desired
                    #
                    if progress:
                        sys.stdout.write('\r')
                        sys.stdout.write("Reading '%s' from '%s': block %d from %d" \
                                      % (vname, fname, nb, nl))
                        sys.stdout.flush()

    if (progress):
      sys.stdout.write('\n')
      sys.stdout.flush()

    if progress:
        sys.stdout.write("Populating AMR structure\n")
    base.populateNodeData()

    if progress:
        sys.stdout.write("Generating OverlappingAMR VTK files\n")

    ofile = "{}_{:06d}.vthb".format(vname, nr)
    opath = "{}_{:06d}".format(vname, nr)
    if not os.path.exists(opath):
        os.makedirs(opath)
    with open(ofile, 'w') as vtk:
        vtk.write('<VTKFile type="vtkOverlappingAMR" version="1.1" ' + \
                    'byte_order="LittleEndian" header_type="UInt64">\n')
        vtk.write('  <vtkOverlappingAMR ' + \
                    'origin="{} {} {}" '.format(*base.lower) + \
                    'grid_description="XYZ">\n')

        fmt = '{}_{:0' + str(len(str(ml))) + '}_{:0' + str(len(str(base.nodes))) + 'd}.vti'

        m = 0
        for lv in range(ml):

            cw = base.size / (rm * nn * 2**lv)
            vtk.write('    <Block level="{}"'.format(lv) + \
                            ' spacing="{} {} {}">\n'.format(*cw))

            no = 0
            for item in base.getNodesFromLevel(lv):
                lo = np.array(item.index) * bm
                up = lo + bm - 1
                ll = np.stack((lo,up)).T.flatten()
                if item.hasData:
                    vfile = os.path.join(opath, fmt.format(vname, lv, no))
                    WriteVTK(vfile, label, item.data, \
                        origin = (item.lower[0], item.lower[1], item.lower[2]), \
                        spacing = (cw[0], cw[1], cw[2]), \
                        compression=compression, compression_level=compression_level)
                    vtk.write('      <DataSet index="{}"'.format(no) + \
                                ' amr_box = "{} {} {} {} {} {}"'.format(*ll) + \
                                    ' file = "{}"></DataSet>\n'.format(vfile))
                    no += 1
                else:
                    vtk.write('      <DataSet index="{}"'.format(no) + \
                                ' amr_box = "{} {} {} {} {} {}"'.format(*ll) + \
                                    '></DataSet>\n')
                m += 1

                if progress:
                    sys.stdout.write('\r')
                    sys.stdout.write("Storing AMR block {} from {}".format(m, base.nodes))
                    sys.stdout.flush()

            vtk.write('    </Block>\n')

        vtk.write('  </vtkOverlappingAMR>\n')
        vtk.write('</VTKFile>')

    if (progress):
        sys.stdout.write('\n')
        sys.stdout.flush()



"""
================================================================================

  This file is part of the AMUN source code, a program to perform
  Newtonian or relativistic magnetohydrodynamical simulations on uniform or
  adaptive mesh.

  Copyright (C) 2018-2022 Grzegorz Kowal <grzegorz@amuncode.org>

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

================================================================================

 submodule: octree.py

  Python module to handle the octree structure of blocks used by
  the AMUN's code adaptive mesh representation of the domain.

  This submodule provides two classes:
   OcBase - the base of the structure
   OcNode - the node can be any node with or without data

  The only requirements for this package are:

     - numpy

--------------------------------------------------------------------------------
"""
import itertools

class OcBase(object):
    ''' Octree base '''

    def __init__(self, position, size, dims):
        self.level    = -1
        self.lower    = position
        self.upper    = [p + s for p, s in zip(position, size)]
        self.size     = size
        self.dims     = dims
        self.nodes    = 0
        self.children = []

        blkSize = [s / d for s, d in zip(size, dims)]

        for k, j, i in itertools.product(range(dims[2]), range(dims[1]), range(dims[0])):
            idx = [i, j, k]
            pos = [l + i * s for l, i, s in zip(self.lower, idx, blkSize)]
            self.children.append(OcNode(idx, pos, blkSize, self.level + 1))
            self.nodes += 1


    ''' Function creates the branch of leafs at given position down to the given level'''
    def createNodeBranch(self, position, level):

        for n in range(len(self.children)):
            self.nodes += self.children[n].createNodeBranch(position, level)


    ''' Function assigns data to the node '''
    def setNodeData(self, position, level, data):

        for n in range(len(self.children)):
            self.children[n].setNodeData(position, level, data)


    ''' Function populates all nodes at lower levels with data from higher levels '''
    def populateNodeData(self):

        for n in range(len(self.children)):
            self.children[n].populateNodeData()


    ''' Function returns the list of nodes from the given level '''
    def getNodesFromLevel(self, level):

        items = []

        for n in range(len(self.children)):
            items = items + self.children[n].getNodesFromLevel(level)

        return items


    ''' Function prints the node properties '''
    def printNodes(self, l):

        print(self.lower, self.upper, self.size, self.level)

        for n in range(len(self.children)):
            self.children[n].printNodes(1, l)



class OcNode(object):
    ''' Octree node '''

    def __init__(self, index, position, size, level):
        self.isLeaf   = True
        self.hasData  = False
        self.level    = level
        self.index    = index
        self.lower    = position
        self.upper    = [p + s for p, s in zip(position, size)]
        self.size     = size
        self.children = []
        self.data     = None


    ''' Functions creates the subnodes '''
    def __refine(self):

        newnodes = 0

        if self.isLeaf:

            newSize  = ( self.size[0] / 2, self.size[1] / 2, self.size[2] / 2 )

            for k, j, i in itertools.product(range(2), range(2), range(2)):

                idx = [i, j, k]
                pos = [l + i * s for l, i, s in zip(self.lower, idx, newSize)]
                idx = [2 * l + i for l, i in zip(self.index, idx)]

                self.children.append(OcNode(idx, pos, newSize, self.level + 1))

                newnodes += 1

            self.isLeaf = False

        return newnodes


    ''' Function creates the branch of leafs at given position down to the given level'''
    def createNodeBranch(self, position, level):
        flag = [(l < p and p < u) \
               for l, p, u in zip(self.lower, position, self.upper)]

        newnodes = 0

        if all(flag):
            if self.level < level:
                if self.isLeaf:
                    newnodes = self.__refine()
                for n in range(len(self.children)):
                    newnodes += self.children[n].createNodeBranch(position, level)

        return newnodes


    ''' Function assigns data to the node '''
    def setNodeData(self, position, level, data):

        flag = [(l < p and p < u) \
               for l, p, u in zip(self.lower, position, self.upper)]

        if all(flag):
            if self.isLeaf:
                self.data = data
                self.hasData = True
            else:
                for n in range(len(self.children)):
                    self.children[n].setNodeData(position, level, data)


    ''' Function populates all nodes at lower levels with data from higher levels '''
    def populateNodeData(self):
        from .interpolation import rebin
        import numpy

        if not self.isLeaf:
            for n in range(len(self.children)):
                self.children[n].populateNodeData()

            if isinstance(self.children[0].data, (list, tuple)):

                self.data = []

                for m, comp in enumerate(self.children[0].data):

                    bm  = comp.shape
                    dm  = [ 2 * d for d in bm ]

                    arr = numpy.zeros(dm, dtype=comp.dtype)

                    for k, j, i in itertools.product(range(2), range(2), range(2)):
                        n = (k * 2 + j) * 2 + i
                        ib, jb, kb =  0 + i * bm[0],  0 + j * bm[1],  0 + k * bm[2]
                        ie, je, ke = ib +     bm[0], jb +     bm[1], kb +     bm[2]
                        arr[kb:ke,jb:je,ib:ie] = self.children[n].data[m][:,:,:]

                    self.data.append(rebin(arr, bm))


            else:
                bm  = self.children[0].data.shape
                dm  = [ 2 * d for d in bm ]

                arr = numpy.zeros(dm, dtype=self.children[0].data.dtype)

                for k, j, i in itertools.product(range(2), range(2), range(2)):
                    n = (k * 2 + j) * 2 + i
                    ib, jb, kb =  0 + i * bm[0],  0 + j * bm[1],  0 + k * bm[2]
                    ie, je, ke = ib +     bm[0], jb +     bm[1], kb +     bm[2]
                    arr[kb:ke,jb:je,ib:ie] = self.children[n].data[:,:,:]

                self.data = rebin(arr, bm)

            self.hasData = True


    ''' Function returns the list of nodes from the given level '''
    def getNodesFromLevel(self, level):

        items = []

        if self.level == level:
            items.append(self)
        elif self.level < level:
            for n in range(len(self.children)):
                items = items + self.children[n].getNodesFromLevel(level)

        return items


    ''' Function prints the node properties '''
    def printNodes(self, s, l):

        if self.level <= l:
            print(s*'  ', "[{:6.3f} {:6.3f} {:6.3f}]".format(*self.lower), \
                    "-> [{:6.3f} {:6.3f} {:6.3f}]".format(*self.upper), \
                    ", size: [{:6.3f} {:6.3f} {:6.3f}]".format(*self.size), \
                    ", level: {}".format(self.level), \
                    ", leaf: {}, data: {}".format(self.isLeaf, self.hasData))

            for n in range(len(self.children)):
                self.children[n].printNodes(s+1, l)



"""
================================================================================

  This file is part of the AMUN source code, a program to perform
  Newtonian or relativistic magnetohydrodynamical simulations on uniform or
  adaptive mesh.

  Copyright (C) 2018-2022 Grzegorz Kowal <grzegorz@amuncode.org>

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

================================================================================

 module: INTERPOLATION

  Support module for Amun snapshots for the block prolongation with different
  types of interpolation.

--------------------------------------------------------------------------------
"""
try:
    from scipy.ndimage import zoom
    from scipy.interpolate import splrep, splev, interp1d, pchip_interpolate
    scipy_available = True
except ImportError:
    scipy_available = False


def rebin(a, newshape):
    '''
        Subroutine changes the size of the input array to to new shape,
        by copying cells or averaging them.
    '''
    import numpy

    assert len(a.shape) == len(newshape)

    m = a.ndim - 1
    if (a.shape[m] > newshape[m]):
        if a.ndim == 3:
            nn = [newshape[0], a.shape[0] // newshape[0], \
                  newshape[1], a.shape[1] // newshape[1], \
                  newshape[2], a.shape[2] // newshape[2]]
            return a.reshape(nn).mean(5).mean(3).mean(1)
        else:
            nn = [newshape[0], a.shape[0] // newshape[0], \
                  newshape[1], a.shape[1] // newshape[1]]
            return a.reshape(nn).mean(3).mean(1)
    else:
        for n in range(a.ndim):
            a = numpy.repeat(a, newshape[n] // a.shape[n], axis=n)
        return(a)


def interpolate(a, newshape, nghosts=0, method=None, order=1):
    '''
        Subroutine rescales the block by interpolating its values.
    '''
    import numpy

    if method == None or method == 'rebin' or not scipy_available:

        ng = nghosts
        if a.ndim == 3:
            return rebin(a[ng:-ng,ng:-ng,ng:-ng], newshape)
        else:
            return rebin(a[ng:-ng,ng:-ng], newshape)

    elif method == 'zoom':

        zf = (newshape[1] // (a.shape[1] - 2 * nghosts))
        ng = zf * nghosts
        if a.ndim == 3:
            return zoom(a, zf, order=order, grid_mode=True, mode='nearest')[ng:-ng,ng:-ng,ng:-ng]
        else:
            return zoom(a, zf, order=order, grid_mode=True, mode='nearest')[ng:-ng,ng:-ng]

    elif method in [ 'monotonic', 'pchip' ]:

        dims = numpy.arange(a.ndim)
        q = a

        for n in dims:

            d2 = numpy.roll(q,-1, axis=0) + numpy.roll(q, 1, axis=0) - 2.0 * q
            q  = q - d2 / 24.0

            d  = numpy.array(q.shape)

            xo = (numpy.arange(0.5, a.shape[n]) - nghosts) / (a.shape[n] - 2 * nghosts)
            xn =  numpy.arange(0.5, newshape[n]) / newshape[n]

            u = q.reshape([d[0], q.size // d[0]])
            f = numpy.zeros([newshape[n], q.size // d[0]])
            for i in range(q.size // d[0]):
                f[:,i] = pchip_interpolate(xo, u[:,i], xn)

            d[0] = newshape[n]
            f = f.reshape(d)

            q = f.transpose(numpy.roll(dims, -1))

        return q

    elif method == 'spline':

        dims = numpy.arange(a.ndim)
        q = a

        for n in dims:

            d2 = numpy.roll(q,-1, axis=0) + numpy.roll(q, 1, axis=0) - 2.0 * q
            q  = q - d2 / 24.0

            d  = numpy.array(q.shape)

            xo = (numpy.arange(0.5, a.shape[n]) - nghosts) / (a.shape[n] - 2 * nghosts)
            xn =  numpy.arange(0.5, newshape[n]) / newshape[n]

            u = q.reshape([d[0], q.size // d[0]])
            f = numpy.zeros([newshape[n], q.size // d[0]])
            for i in range(q.size // d[0]):
                t = splrep(xo, u[:,i], k=5, s=0.0)
                f[:,i] = splev(xn, t)

            d[0] = newshape[n]
            f = f.reshape(d)

            q = f.transpose(numpy.roll(dims, -1))

        return q

    else:

        dims = numpy.arange(a.ndim)
        q = a

        for n in dims:

            d2 = numpy.roll(q,-1, axis=0) + numpy.roll(q, 1, axis=0) - 2.0 * q
            q  = q - d2 / 24.0

            d  = numpy.array(q.shape)

            xo = (numpy.arange(0.5, a.shape[n]) - nghosts) / (a.shape[n] - 2 * nghosts)
            xn =  numpy.arange(0.5, newshape[n]) / newshape[n]

            u = q.reshape([d[0], q.size // d[0]])
            f = numpy.zeros([newshape[n], q.size // d[0]])
            for i in range(q.size // d[0]):
                t = interp1d(xo, u[:,i], kind=method)
                f[:,i] = t(xn)

            d[0] = newshape[n]
            f = f.reshape(d)

            q = f.transpose(numpy.roll(dims, -1))

        return q



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="amunpy",
    version="0.9.6",
    author="Grzegorz Kowal",
    author_email="grzegorz@amuncode.org",
    description="Python Interface for the AMUN code's snapshots",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://www.amuncode.org/",
    project_urls={
        "Bug Tracker": "https://bitbucket.org/amunteam/amun-code/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.6",
    install_requires=['h5py', 'numpy', 'xxhash', 'lz4', 'zstandard'],
    extras_require={
        "interpolation": ['scipy'],
    }
)
